# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for writing CSV formatted annotations."""
import logging
import os
from typing import List, Optional, Tuple

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_writer import AnnotationWriter
from mlcvzoo_base.api.data.dataset_info import BaseDatasetInfo
from mlcvzoo_base.configuration.annotation_handler_config import (
    AnnotationHandlerWriteOutputConfig,
    AnnotationHandlerWriteOutputCSVAnnotationConfig,
)
from mlcvzoo_base.configuration.structs import (
    CSVFileNameIdentifier,
    FileNamePlaceholders,
)
from mlcvzoo_base.data_preparation.structs import CSVOutputStringFormats
from mlcvzoo_base.data_preparation.utils import replace_dir_by_index, save_bbox_snippets
from mlcvzoo_base.utils.annotation_utils import (
    create_annotation_file_from_list,
    create_cross_val_list_splits,
    create_list_split,
)
from mlcvzoo_base.utils.common_utils import get_current_timestamp

logger = logging.getLogger(__name__)


class CSVAnnotationWriter(AnnotationWriter):
    """
    Writer for generating a csv file based on a given
    list of annotations
    """

    def __init__(
        self,
        write_output_config: AnnotationHandlerWriteOutputConfig,
        output_string_format: str = CSVOutputStringFormats.BASE,
    ) -> None:
        self.csv_base_file_name: Optional[str] = None
        self.write_output_config = write_output_config
        self.output_string_format = output_string_format

        # define attribute types
        self.list_splits: List[Tuple[List[str], BaseDatasetInfo]]

    def write(self, annotations: List[BaseAnnotation]) -> Optional[str]:
        # TODO: Control the filename of the csv and avoid that it is different based on date and time
        output_file_path: Optional[str] = None

        assert self.write_output_config.csv_annotation is not None

        self.list_splits = self.__generate_csv_lists_from_annotations(
            annotations=annotations, output_string_format=self.output_string_format
        )

        timestamp_string = get_current_timestamp(
            self.write_output_config.csv_annotation.timestamp_format
        )

        file_extension = CSVAnnotationWriter.__get_file_extension(
            output_string_format=self.output_string_format
        )

        self.__replace_file_name_placeholders(timestamp_string=timestamp_string)

        csv_base_file_name = CSVAnnotationWriter.get_csv_base_file_name(
            timestamp_string=timestamp_string,
            csv_annotation=self.write_output_config.csv_annotation,
        )

        if csv_base_file_name is not None:
            self.csv_base_file_name = csv_base_file_name
        else:
            raise ValueError("Could find a correct setting for 'csv_base_file_name'")

        # TODO: Select between more than cross validation and just split?
        # CROSS VALIDATION
        if self.write_output_config.use_cross_val:
            for index, list_split in enumerate(self.list_splits):
                # TODO: make output_path configurable via the configuration class
                output_file_path = os.path.join(
                    self.write_output_config.csv_annotation.csv_dir,
                    f"{csv_base_file_name}_"
                    f"{CSVFileNameIdentifier.CROSS_VAL_SPLIT}_"
                    f"{index}{file_extension}",
                )

                create_annotation_file_from_list(
                    csv_entry_list=list_split[0],
                    dataset_info=list_split[1],
                    output_file_path=output_file_path,
                )

        # TRAIN / EVAL / VALIDATION Split
        else:
            # TRAIN
            if self.write_output_config.split_size > 0.0:
                output_file_path = CSVAnnotationWriter.generate_csv_path(
                    csv_base_file_name=csv_base_file_name,
                    file_extension=file_extension,
                    file_identifier=CSVFileNameIdentifier.TRAINING,
                    csv_annotation=self.write_output_config.csv_annotation,
                )
            # VALIDATION
            else:
                output_file_path = CSVAnnotationWriter.generate_csv_path(
                    csv_base_file_name=csv_base_file_name,
                    file_extension=file_extension,
                    file_identifier=CSVFileNameIdentifier.VALIDATION,
                    csv_annotation=self.write_output_config.csv_annotation,
                )

            create_annotation_file_from_list(
                csv_entry_list=self.list_splits[0][0],
                dataset_info=self.list_splits[0][1],
                output_file_path=output_file_path,
            )

        # EVALUATION
        if len(self.list_splits) == 2:
            output_file_path = CSVAnnotationWriter.generate_csv_path(
                csv_base_file_name=csv_base_file_name,
                file_extension=file_extension,
                file_identifier=CSVFileNameIdentifier.EVALUATION,
                csv_annotation=self.write_output_config.csv_annotation,
            )

            create_annotation_file_from_list(
                csv_entry_list=self.list_splits[1][0],
                dataset_info=self.list_splits[1][1],
                output_file_path=output_file_path,
            )

        return output_file_path

    def __annotation_to_csv_entry_list(
        self, annotations: List[BaseAnnotation], output_string_format: str
    ) -> Tuple[List[str], BaseDatasetInfo]:
        csv_entry_list: List[str] = list()

        # TODO: do we need this statistic?
        dataset_info = BaseDatasetInfo(
            classes_path="",
            image_count=0.0,
        )

        for annotation in annotations:
            annotation = replace_dir_by_index(annotation=annotation)

            # Write annotation data to csv
            if len(annotation.get_bounding_boxes(include_segmentations=True)) > 0:
                if self.write_output_config.csv_annotation is None:
                    raise ValueError(
                        "In order to write data to a csv the "
                        "write_output_config.csv_annotation configuration attribute"
                        "has to be provided!"
                    )

                csv_entry = annotation.to_csv_entry(
                    include_surrounding_bboxes=(
                        self.write_output_config.csv_annotation.include_surrounding_bboxes
                    ),
                    output_string_format=output_string_format,
                    use_difficult=self.write_output_config.use_difficult,
                    use_occluded=self.write_output_config.use_occluded,
                )

                if csv_entry is None or csv_entry == "":
                    logger.warning("csv_entry is empty, skip ....")
                    continue

                logger.debug("Add csv_entry: %s", csv_entry)

                csv_entry_list.append(csv_entry)

            # Update info-dict statistics with the data of this annotation
            dataset_info.update(annotation=annotation)

            # Cut out bounding-box snippets from original image.
            # This can be useful for classification of the bounding-boxes
            if self.write_output_config.bbox_snippets is not None:
                save_bbox_snippets(
                    annotation=annotation,
                    output_dir=self.write_output_config.bbox_snippets.output_dir,
                )

        return csv_entry_list, dataset_info

    def __generate_csv_lists_from_annotations(
        self, annotations: List[BaseAnnotation], output_string_format: str
    ) -> List[Tuple[List[str], BaseDatasetInfo]]:
        """
        Creates a list of data sets which are splits from the given annotations.

        Args:
            annotations: a list of BaseAnnotation objects
            output_string_format: string, one of CSVOutputStringFormats

        Returns: a list of data and respective dataset info

        """

        list_splits: List[Tuple[List[str], BaseDatasetInfo]] = list()

        # CROSS VALIDATION
        if self.write_output_config.use_cross_val:
            cross_val_list_splits = create_cross_val_list_splits(
                input_file_list=annotations,
                number_splits=self.write_output_config.number_splits,
            )

            for cross_val_list_split in cross_val_list_splits:
                csv_entry_list, dataset_info = self.__annotation_to_csv_entry_list(
                    annotations=cross_val_list_split,
                    output_string_format=output_string_format,
                )

                list_splits.append((csv_entry_list, dataset_info))

        # TRAIN / EVAL / VALIDATION Split
        else:
            # TRAIN / EVAL
            if self.write_output_config.split_size > 0.0:
                train_list, eval_list = create_list_split(
                    input_file_list=annotations,
                    split_size=self.write_output_config.split_size,
                    random_state=self.write_output_config.random_state,
                )

                (train_csv_entry_list, train_dataset_info,) = self.__annotation_to_csv_entry_list(
                    annotations=train_list, output_string_format=output_string_format
                )

                (eval_csv_entry_list, eval_dataset_info,) = self.__annotation_to_csv_entry_list(
                    annotations=eval_list, output_string_format=output_string_format
                )

                list_splits.append((train_csv_entry_list, train_dataset_info))
                list_splits.append((eval_csv_entry_list, eval_dataset_info))

            # VALIDATION
            else:
                (train_csv_entry_list, train_dataset_info,) = self.__annotation_to_csv_entry_list(
                    annotations=annotations, output_string_format=output_string_format
                )

                list_splits.append((train_csv_entry_list, train_dataset_info))

        return list_splits

    def __replace_file_name_placeholders(self, timestamp_string: str) -> None:
        assert self.write_output_config.csv_annotation is not None

        # TODO: add test
        if (
            FileNamePlaceholders.TIMESTAMP
            in self.write_output_config.csv_annotation.csv_split_0_file_name
        ):
            self.write_output_config.csv_annotation.csv_split_0_file_name = (
                self.write_output_config.csv_annotation.csv_split_0_file_name.replace(
                    FileNamePlaceholders.TIMESTAMP, timestamp_string
                )
            )

        # TODO: add test
        if (
            FileNamePlaceholders.TIMESTAMP
            in self.write_output_config.csv_annotation.csv_split_1_file_name
        ):
            self.write_output_config.csv_annotation.csv_split_1_file_name = (
                self.write_output_config.csv_annotation.csv_split_1_file_name.replace(
                    FileNamePlaceholders.TIMESTAMP, timestamp_string
                )
            )

    @staticmethod
    def get_csv_base_file_name(
        timestamp_string: str,
        csv_annotation: Optional[AnnotationHandlerWriteOutputCSVAnnotationConfig],
    ) -> Optional[str]:
        """
        Get a csv basis file name based on a timestamp and given configuration.

        Args:
            timestamp_string: timestamp as string to identify a training process
            csv_annotation: the csv-annotation config object

        Returns:
            Optional, base filename of csv files that hold information about
            training, validation and evaluation
        """

        csv_base_file_name: Optional[str] = None

        if csv_annotation is not None:
            if FileNamePlaceholders.TIMESTAMP in csv_annotation.csv_base_file_name:
                csv_base_file_name = csv_annotation.csv_base_file_name.replace(
                    FileNamePlaceholders.TIMESTAMP, timestamp_string
                )
            else:
                csv_base_file_name = csv_annotation.csv_base_file_name

        return csv_base_file_name

    @staticmethod
    def generate_csv_path(
        csv_base_file_name: Optional[str],
        file_extension: str,
        file_identifier: str,
        csv_annotation: Optional[AnnotationHandlerWriteOutputCSVAnnotationConfig] = None,
    ) -> str:
        """
        Generates a path to a csv file where annotation information is stored.

        Args:
            csv_base_file_name: Optional name of the csv file with training data
            file_extension: string, file extension (format)
            file_identifier: string, data set identifier (one of CSVFileNameIdentifier)
            csv_annotation: Optional configuration for output formatting

        Returns: string, the annotation output file path

        """

        if csv_annotation is not None:
            if csv_base_file_name is not None:
                train_file_name = f"{csv_base_file_name}_" f"{file_identifier}" f"{file_extension}"
            elif csv_annotation.csv_split_0_file_name != "":
                train_file_name = csv_annotation.csv_split_0_file_name
            else:
                train_file_name = f"{file_identifier}{file_extension}"

            output_file_path = os.path.join(
                csv_annotation.csv_dir,
                train_file_name,
            )

        else:
            if csv_base_file_name is not None:
                output_file_path = (
                    f"{csv_base_file_name}_" f"{file_identifier}" f"{file_extension}"
                )
            else:
                output_file_path = f"{file_identifier}" f"{file_extension}"

        return output_file_path

    @staticmethod
    def __get_file_extension(output_string_format: str) -> str:
        if output_string_format == CSVOutputStringFormats.BASE:
            file_extension = ".csv"
        elif output_string_format == CSVOutputStringFormats.YOLO:
            file_extension = ".txt"
        else:
            raise ValueError(
                f"file-format '{output_string_format}' is not valid to "
                f"be used for annotation file generation."
                f"Please use either of "
                f"{CSVOutputStringFormats.get_values_as_string(class_type=CSVOutputStringFormats)}"
            )

        return file_extension



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

import logging
from typing import Dict, List, Optional

import related
from attr import define
from config_builder import BaseConfigClass

from mlcvzoo_base.configuration.annotation_handler_config import AnnotationHandlerConfig
from mlcvzoo_base.configuration.mlfow_config import MLFlowConfig
from mlcvzoo_base.configuration.model_config import ModelConfig
from mlcvzoo_base.configuration.visualization_config import VisualizationConfig

logger = logging.getLogger(__name__)


@define
class ODEvaluationInputDataConfig(BaseConfigClass):
    # path to the csv file, where the annotation data should be parsed from
    csv_file_path: str = related.StringField()


@define
class TensorboardLoggingConfig(BaseConfigClass):
    tensorboard_dir: str = related.StringField()
    false_positive_image_size: int = related.IntegerField(default=650)


@define
class ODEvaluationConfig(BaseConfigClass):
    iou_thresholds: List[float] = related.SequenceField(float)

    mlflow_config: Optional[MLFlowConfig] = related.ChildField(
        cls=MLFlowConfig, required=False, default=None
    )

    tensorboard_logging: Optional[TensorboardLoggingConfig] = related.ChildField(
        cls=TensorboardLoggingConfig, required=False, default=None
    )

    visualization: VisualizationConfig = related.ChildField(
        cls=VisualizationConfig, default=VisualizationConfig()
    )

    model_configs: Optional[List[ModelConfig]] = related.SequenceField(
        cls=ModelConfig, required=False, default=None
    )

    input_data: Optional[List[ODEvaluationInputDataConfig]] = related.SequenceField(
        cls=ODEvaluationInputDataConfig, required=False, default=None
    )

    annotation_handler_config: Optional[AnnotationHandlerConfig] = related.ChildField(
        cls=AnnotationHandlerConfig, required=False, default=None
    )

    def check_values(self) -> bool:
        logger.warning(
            "DEPRECATED: The class ODEvaluationConfig is deprecated and will"
            "be removed in future versions"
        )
        return True



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for parsing annotations"""
from abc import ABC, abstractmethod
from typing import List

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper


class AnnotationParser(ABC):
    """
    Super class for defining the methods that are needed to parse a list of
    instances that are of the type BaseAnnotation.
    Each annotation format e.g. Pascal-VOC, COCO, CVAT-for-images should have
    its own child AnnotationHandler class
    """

    csv_directory_replacement_string = "IMAGE_DIR_{}"

    def __init__(self, mapper: AnnotationClassMapper):
        self.mapper: AnnotationClassMapper = mapper

    @abstractmethod
    def parse(self) -> List[BaseAnnotation]:
        """
        Parses annotations in respective format to BaseAnnotations

        Returns: a List of BaseAnnotations
        """

        raise NotImplementedError()



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for different utility operations regarding configuration operations."""
import logging
import os
from argparse import ArgumentTypeError
from typing import Dict, Optional, Tuple, Type, Union, cast

import yaml
from attrs import asdict
from config_builder import BaseConfigClass, ConfigBuilder
from config_builder.yaml_constructors import (
    join_object,
    join_object_from_config_dir,
    join_path,
    join_string,
    join_string_with_delimiter,
)
from related import to_model, to_yaml

from mlcvzoo_base.api.configuration import ModelConfiguration
from mlcvzoo_base.configuration.replacement_config import ReplacementConfig

logger = logging.getLogger(__name__)


def get_replacement_map_from_replacement_config(
    yaml_config_path: str,
) -> Tuple[Dict[str, str], Optional[ReplacementConfig]]:
    """
    Convenient method to fill the (global) replacement map of the config-builder.
    In the MLCVZoo all attributes of the ReplacementConfig class can be used as
    replacements in string configuration-attributes, as well as replacements in
    config-paths which are defined in e.g. !join_object. This Method parses the
    ReplacementConfig given by the yaml_config_path parameter and sets the corresponding
    values for the replacement map in the config-builder according to the parsed content.

    Args:
        yaml_config_path: The config-file where to parse the ReplacementConfig from

    Returns:
        Dictionary with a sting mapping of placeholders to actual os paths
    """
    replacement_config: Optional[ReplacementConfig] = None

    string_replacement_map: Dict[str, str] = {}

    if os.path.isfile(yaml_config_path):
        replacement_config = parse_replacement_config(yaml_config_path=yaml_config_path)

        logger.info(
            "Successfully extracted content for placeholders "
            "from the replacement config given by: \n"
            " - yaml_config_path: %s\n"
            " - replacement_config: %s\n\n"
            f"Use these placeholders to update the string replacement map\n"
            % (yaml_config_path, replacement_config)
        )

        for key, value in replacement_config.to_dict().items():
            if isinstance(value, str):
                string_replacement_map[key] = value
    else:
        logger.warning(
            f"The given yaml_config_path '{yaml_config_path} does not exist!\n"
            "No content for placeholders could be extracted from "
            f"the replacement config given by: {yaml_config_path}. \n"
            f"Only update placeholder identifier from the replacement config (empty values)"
            f": {list(ReplacementConfig.__dict__['__annotations__'].keys())}"
        )

        for key, value in ReplacementConfig.__dict__["__annotations__"].items():
            if isinstance(value, str):
                string_replacement_map[key] = ""

    return string_replacement_map, replacement_config


def parse_replacement_config(
    yaml_config_path: str, replacement_config_key: str = "replacement_config"
) -> ReplacementConfig:
    """
    Parse an ReplacementConfig object from the given yaml_config_path.

    Args:
        yaml_config_path: string, path to a yaml configuration file
        replacement_config_key: string, key for replacement config in YAML

    Returns:
        The ReplacementConfig object derived from the file
    """
    # register the tag handlers
    yaml.add_constructor("!join_string", join_string)
    yaml.add_constructor("!join_string_with_delimiter", join_string_with_delimiter)
    yaml.add_constructor("!join_path", join_path)
    yaml.add_constructor("!join_object", join_object)
    yaml.add_constructor("!join_object_from_config_dir", join_object_from_config_dir)

    with open(file=yaml_config_path, mode="r", encoding="utf8") as config_file:
        yaml_dict = yaml.load(stream=config_file, Loader=yaml.Loader)

    if replacement_config_key in yaml_dict:
        replacement_config_dict = yaml_dict.pop(replacement_config_key)
    else:
        replacement_config_dict = yaml_dict

    return cast(ReplacementConfig, to_model(ReplacementConfig, replacement_config_dict))


def str2bool(v: Union[bool, str]) -> bool:
    """
    Parses an object to a boolean value.

    Args:
        v: Any, an expression which indicates a boolean decision
            so either a bool variable or
            a string with one of "yes", "true", "t", "y", "1" or
            one of "no", "false", "f", "n", "0"

    Returns:
         bool, the closest boolean equivalent to the given input

    Raises:
        ArgumentTypeError if no reasonable matching is possible
    """
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1", "1.0"):
        return True
    if v.lower() in ("no", "false", "f", "n", "0", "0.0"):
        return False
    raise ArgumentTypeError("Boolean value expected.")


def create_configuration(
    configuration_class: Type[ModelConfiguration],
    from_yaml: Optional[str] = None,
    input_configuration: Optional[ModelConfiguration] = None,
    string_replacement_map: Optional[Dict[str, str]] = None,
) -> ModelConfiguration:
    """
    Utility method to build a configuration object.

    Args:
        configuration_class: The type of the configuration object
        from_yaml: (Optional) A yaml filepath where to build the configuration
                    object from
        input_configuration: (Optional) An already existing configuration object
        string_replacement_map: A dictionary that defines placeholders which can be used
                                while parsing the file. They can be understood as variables
                                that can be used to define configs that are valid across
                                multiple devices.

    Returns:
        The created configuration object
    """

    logger.info(
        f"\n==========================================\n"
        f"Create configuration:\n"
        f" - class {configuration_class} \n"
        f" - from yaml: {from_yaml}\n"
        f" - input configuration: {input_configuration}\n"
        f"=========================================="
    )

    if not input_configuration:
        configuration: ModelConfiguration = cast(
            ModelConfiguration,
            ConfigBuilder(
                class_type=configuration_class,
                yaml_config_path=from_yaml,
                string_replacement_map=string_replacement_map,
            ).configuration,
        )
    else:
        configuration = input_configuration

    return configuration


def write_config_to_yaml(config: BaseConfigClass, output_yaml_config_path: str) -> None:
    """

    Args:
        config:
        output_yaml_config_path:

    Returns:

    """

    with open(file=output_yaml_config_path, mode="w") as output_yaml_file:
        logger.info("Write current model-configuration to file: %s", output_yaml_config_path)

        to_yaml(
            obj=config,
            yaml_package=yaml,
            dumper_cls=yaml.Dumper,
            stream=output_yaml_file,
        )



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for parsing MOT formatted annotations"""

import logging
import os
from typing import Dict, List

import cv2

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.annotation_parser import AnnotationParser
from mlcvzoo_base.configuration.annotation_handler_config import (
    AnnotationHandlerMOTInputDataConfig,
)
from mlcvzoo_base.configuration.class_mapping_config import (
    ClassMappingConfig,
    ClassMappingModelClassesConfig,
)
from mlcvzoo_base.data_preparation.annotation_builder.mot_annotation_builder import (
    MOTAnnotationBuilder,
)
from mlcvzoo_base.data_preparation.structs import MOTChallengeFormats
from mlcvzoo_base.utils.file_utils import get_file_list

logger = logging.getLogger(__name__)


class MOTAnnotationParser(AnnotationParser):
    """
    Class for parsing annotations in MOT challenge format.
    Please see MOTChallengeFormats for supported formats.
    """

    def __init__(
        self,
        mapper: AnnotationClassMapper,
        mot_input_data: List[AnnotationHandlerMOTInputDataConfig],
    ) -> None:
        AnnotationParser.__init__(self, mapper=mapper)

        self.mot_input_data = mot_input_data

    @staticmethod
    def create_mot_mapper(mot_format: str) -> AnnotationClassMapper:
        if mot_format == MOTChallengeFormats.MOT20.value:
            mot_class_mapper = AnnotationClassMapper(
                class_mapping=MOTAnnotationParser.create_mot__2020_class_mapping()
            )
        elif mot_format == MOTChallengeFormats.MOT1617.value:
            mot_class_mapper = AnnotationClassMapper(
                class_mapping=MOTAnnotationParser.create_mot__201617_class_mapping()
            )
        elif mot_format == MOTChallengeFormats.MOT15.value:
            mot_class_mapper = AnnotationClassMapper(
                class_mapping=MOTAnnotationParser.create_mot__2015_class_mapping()
            )
        else:
            raise ValueError(
                f"mot_format='{mot_format}' not valid. "
                f"Please use one of "
                f"{[f.value for f in MOTChallengeFormats]}"
            )

        return mot_class_mapper

    @staticmethod
    def create_mapper_from_label_file(label_path: str) -> AnnotationClassMapper:

        with open(label_path) as label_file:
            label_lines = label_file.readlines()

        # MOT class IDs start with 1
        class_mapping_configs: List[ClassMappingModelClassesConfig] = [
            ClassMappingModelClassesConfig(class_name="DEFAULT", class_id=0)
        ]
        for class_id, label_line in enumerate(label_lines):
            class_name = label_line.strip()

            if class_name != "":
                class_mapping_configs.append(
                    ClassMappingModelClassesConfig(class_name=class_name, class_id=class_id + 1)
                )

        return AnnotationClassMapper(
            class_mapping=ClassMappingConfig(
                mapping=[],
                model_classes=class_mapping_configs,
                number_model_classes=len(class_mapping_configs),
            )
        )

    def parse(self) -> List[BaseAnnotation]:
        """
        Parse a list of annotations from a MOT annotation file.
        Details of parsing the different formats can be found in
        mlcvzoo_base.data_preparation.annotation_builder.mot_annotation_builder

        Returns:
            A list of BaseAnnotation objects
        """

        annotations: List[BaseAnnotation] = list()

        for dataset_count, input_data in enumerate(self.mot_input_data):

            if input_data.label_path is not None:
                mot_class_mapper = MOTAnnotationParser.create_mapper_from_label_file(
                    label_path=input_data.label_path
                )
            else:
                mot_class_mapper = MOTAnnotationParser.create_mot_mapper(
                    mot_format=input_data.mot_format
                )

            with open(
                file=input_data.annotation_path, mode="r", encoding="'utf-8"
            ) as annotation_file:
                annotation_lines = annotation_file.readlines()

            # frame-id: annotation line
            annotation_line_dict: Dict[int, List[str]] = {}
            for annotation_line in annotation_lines:
                frame_id = int(annotation_line.split(",")[0])

                if frame_id not in annotation_line_dict:
                    annotation_line_dict[frame_id] = []

                annotation_line_dict[frame_id].append(annotation_line.strip())

            image_paths = get_file_list(
                input_dir=input_data.image_dir,
                search_subfolders=False,
                file_extension=input_data.image_format,
            )

            image_paths.sort()
            _input_data_annotations: List[BaseAnnotation] = []
            for image_id, image_path in enumerate(image_paths):

                # MOT image IDs start with 1
                mot_image_id = image_id + 1
                if mot_image_id not in annotation_line_dict:
                    continue

                image = cv2.imread(image_path)
                image_shape = image.shape

                replacement_string = AnnotationParser.csv_directory_replacement_string.format(
                    dataset_count
                )

                mot_builder = MOTAnnotationBuilder(
                    image_shape=image_shape,
                    image_id=mot_image_id,
                    annotation_line_dict=annotation_line_dict,
                    mapper=self.mapper,
                    mot_class_mapper=mot_class_mapper,
                    mot_format=input_data.mot_format,
                    ground_truth=input_data.ground_truth,
                )

                try:
                    annotation = mot_builder.build(
                        image_path=image_path,
                        annotation_path=input_data.annotation_path,
                        image_dir=input_data.image_dir,
                        annotation_dir=os.path.dirname(input_data.annotation_path),
                        replacement_string=replacement_string,
                    )

                    _input_data_annotations.append(annotation)

                except ValueError as error:
                    logger.warning(
                        "%s, annotation from path %s will be skipped",
                        str(error),
                        input_data.annotation_path,
                    )
                    continue

            logger.info(
                f"Parsed {len(_input_data_annotations)} "
                f"from mot file '{input_data.annotation_path}'"
            )
            annotations.extend(_input_data_annotations)

        return annotations

    @staticmethod
    def create_mot__2015_class_mapping() -> ClassMappingConfig:
        # REMARK:
        # - The class-ids in the MOT Challenge start a 1, therefore we
        #   add an extra class in order to fulfill our definition of a class-mapping
        # - Due to related package an "Unexpected keyword argument" is raised. For now
        #   this error is ignored, but it will be fixed in future versions of related
        # - No line break are definitely more readable here

        return ClassMappingConfig(
            mapping=[],
            model_classes=[
                ClassMappingModelClassesConfig(class_name="other", class_id=0),
                ClassMappingModelClassesConfig(class_name="pedestrian", class_id=1),
            ],
            number_model_classes=2,
        )

    @staticmethod
    def create_mot__201617_class_mapping() -> ClassMappingConfig:
        # REMARK:
        # - The class-ids in the MOT Challenge start a 1, therefore we
        #   add an extra class in order to fulfill our definition of a class-mapping
        # - Due to related package an "Unexpected keyword argument" is raised. For now
        #   this error is ignored, but it will be fixed in future versions of related
        # - No line break are definitely more readable here

        mot_2015_class_mapping = MOTAnnotationParser.create_mot__2015_class_mapping()

        model_classes = [
            ClassMappingModelClassesConfig(class_name="persononvehicle", class_id=2),
            ClassMappingModelClassesConfig(class_name="car", class_id=3),
            ClassMappingModelClassesConfig(class_name="bicycle", class_id=4),
            ClassMappingModelClassesConfig(class_name="motorbike", class_id=5),
            ClassMappingModelClassesConfig(class_name="nonmotorvehicle", class_id=6),
            ClassMappingModelClassesConfig(class_name="staticperson", class_id=7),
            ClassMappingModelClassesConfig(class_name="distractor", class_id=8),
            ClassMappingModelClassesConfig(class_name="occluder", class_id=9),
            ClassMappingModelClassesConfig(class_name="occluderonground", class_id=10),
            ClassMappingModelClassesConfig(class_name="occluderfull", class_id=11),
            ClassMappingModelClassesConfig(class_name="reflection", class_id=12),
        ]

        model_classes.extend(mot_2015_class_mapping.model_classes)

        return ClassMappingConfig(
            mapping=[],
            model_classes=model_classes,
            number_model_classes=len(model_classes),
        )

    @staticmethod
    def create_mot__2020_class_mapping() -> ClassMappingConfig:
        """
        Creates a class mapping that

        Returns:

        """

        # REMARK:
        # - The class-ids in the MOT Challenge start a 1, therefore we
        #   add an extra class in order to fulfill our definition of a class-mapping
        # - Due to related package an "Unexpected keyword argument" is raised. For now
        #   this error is ignored, but it will be fixed in future versions of related
        # - No line break are definitely more readable here

        mot_201617_class_mapping = MOTAnnotationParser.create_mot__201617_class_mapping()

        model_classes = [
            ClassMappingModelClassesConfig(class_name="crowd", class_id=13),
        ]

        model_classes.extend(mot_201617_class_mapping.model_classes)

        return ClassMappingConfig(
            mapping=[],
            model_classes=model_classes,
            number_model_classes=len(model_classes),
        )



# Copyright 2022 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module to provide a registry for configuration constructors"""

import logging
from typing import Type

from config_builder import BaseConfigClass

from mlcvzoo_base.api.registry import MLCVZooRegistry

logger = logging.getLogger(__name__)


class ConfigRegistry(MLCVZooRegistry[Type[BaseConfigClass]]):

    """
    Class to provide a registry for configuration constructors
    """

    def __init__(self) -> None:
        MLCVZooRegistry.__init__(self)



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

from .draw_utils import (
    draw_bbox_cv2,
    draw_on_image,
    draw_on_pil_image,
    draw_polygon_cv2,
    generate_detector_colors,
)
from .file_utils import (
    encoding_safe_imread,
    encoding_safe_imwrite,
    ensure_dir,
    get_file_list,
    get_project_path_information,
)



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Gathering of utility methods that are generating metric output in any
kind of form.
"""

import logging
import os
from datetime import datetime
from typing import Dict, List, Optional

import mlflow
from tensorboardX import SummaryWriter

from mlcvzoo_base.configuration.structs import FileNamePlaceholders
from mlcvzoo_base.evaluation.object_detection.configuration import (
    TensorboardLoggingConfig,
)
from mlcvzoo_base.evaluation.object_detection.data_classes import (
    METRIC_DICT_TYPE,
    METRIC_IMAGE_INFO_TYPE,
    ODModelEvaluationMetrics,
)
from mlcvzoo_base.evaluation.object_detection.structs import BBoxSizeTypes
from mlcvzoo_base.evaluation.object_detection.utils import (
    __create_tf_image,
    generate_img_id_map,
    generate_metric_table,
)
from mlcvzoo_base.utils import ensure_dir

logger = logging.getLogger(__name__)


def log_od_metrics_to_mlflow(
    model_specifier: str,
    metrics_dict: METRIC_DICT_TYPE,
    iou_threshold: float,
    score_threshold: Optional[float] = None,
    nms_threshold: Optional[float] = None,
    step: Optional[int] = None,
) -> None:
    """
    Log the object detection metrics for the given iou threshold with mlflow.
    This includes the logging of any metric that is defined by the
    dataclass mlcvzoo.evaluation.object_detection.data_classes.ODMetrics
    for any bounding box size type defined in the class
    mlcvzoo.evaluation.object_detection.structs.BBoxSizeTypes

    Args:
        model_specifier: String identifying a model
        metrics_dict: The metrics dictionary where to take the metrics from
        iou_threshold: The iou threshold for which the metrics should be logged
        score_threshold: Optionally hand over the score threshold with which the
                         metrics have been generated
        nms_threshold: Optionally hand over the nms threshold with which the
                      metrics have been generated
        step: Metric step that is handed over to mlflow

    Returns:
        None
    """

    if mlflow.active_run() is not None:
        mlflow.log_param("algorithm_type", "Object Detection")

        logger.debug("Log mlflow metrics for model '%s'", model_specifier)

        if score_threshold is not None:
            mlflow.log_param(key="score_threshold", value=score_threshold)
        if nms_threshold is not None:
            mlflow.log_param(key="nms_threshold", value=nms_threshold)

        for bbox_size_type in BBoxSizeTypes.get_values_as_list(class_type=BBoxSizeTypes):
            for class_name, od_metric in metrics_dict[iou_threshold][bbox_size_type].items():
                mlflow.log_metric(
                    key=f"{class_name}_TP_{bbox_size_type}",
                    value=od_metric.TP,
                    step=step,
                )
                mlflow.log_metric(
                    key=f"{class_name}_FP_{bbox_size_type}",
                    value=od_metric.FP,
                    step=step,
                )
                mlflow.log_metric(
                    key=f"{class_name}_FN_{bbox_size_type}",
                    value=od_metric.FN,
                    step=step,
                )
                mlflow.log_metric(
                    key=f"{class_name}_COUNT_{bbox_size_type}",
                    value=od_metric.COUNT,
                    step=step,
                )
                mlflow.log_metric(
                    key=f"{class_name}_RC_{bbox_size_type}",
                    value=od_metric.RC,
                    step=step,
                )
                mlflow.log_metric(
                    key=f"{class_name}_PR_{bbox_size_type}",
                    value=od_metric.PR,
                    step=step,
                )
                mlflow.log_metric(
                    key=f"{class_name}_F1_{bbox_size_type}",
                    value=od_metric.F1,
                    step=step,
                )
                mlflow.log_metric(
                    key=f"{class_name}_AP_{bbox_size_type}",
                    value=od_metric.AP,
                    step=step,
                )
    else:
        logger.warning(
            "Can not log metrics with 'log_od_metrics_to_mlflow' for model '%s' "
            "since no mlflow run is active",
            model_specifier,
        )


def log_false_positive_info(
    model_specifier: str,
    metric_image_info_dict: METRIC_IMAGE_INFO_TYPE,
) -> None:
    """
    TODO

    Args:
        model_specifier:
        metric_image_info_dict:

    Returns:
        None
    """

    for class_name, image_dict in metric_image_info_dict.items():
        for image_path, metric_image_info in image_dict.items():
            if metric_image_info.false_positive_annotation is not None:
                bounding_boxes = metric_image_info.false_positive_annotation.get_bounding_boxes(
                    include_segmentations=True
                )

                logger.debug(
                    "\nFALSE POSITIVE for model %s: \n"
                    "  - image-path: '%s' \n"
                    "  - image-annotation: %s \n"
                    "  - bounding_boxes:   %s \n",
                    model_specifier,
                    metric_image_info.false_positive_annotation.image_path,
                    metric_image_info.false_positive_annotation.annotation_path,
                    bounding_boxes,
                )

            if metric_image_info.false_negative_annotation is not None:
                bounding_boxes = metric_image_info.false_negative_annotation.get_bounding_boxes(
                    include_segmentations=True
                )

                logger.debug(
                    "\nFALSE NEGATIVE for model %s: \n"
                    "  - image-path:       "
                    "%s\n"
                    "  - image-annotation: "
                    "%s\n"
                    "  - bounding_boxes:   %s\n",
                    model_specifier,
                    metric_image_info.false_negative_annotation.image_path,
                    metric_image_info.false_negative_annotation.annotation_path,
                    bounding_boxes,
                )


def log_false_positive_info_to_tb(
    model_name: str,
    metric_image_info_dict: METRIC_IMAGE_INFO_TYPE,
    tb_logging_config: TensorboardLoggingConfig,
) -> None:
    """
    Writes evaluation metrics and images to tensorboard directory.

    Args:
        model_name: Name of the model
        metric_image_info_dict: Dictionary mapping of string to a image dictionary
        tb_logging_config: configuration object defining the behavior of logging the
                           false positive information to tensorboard

    Returns:
        None
    """

    timestamp = datetime.now().strftime("%Y-%m-%dT_%H-%M")

    tb_dir = os.path.join(
        tb_logging_config.tensorboard_dir.replace(FileNamePlaceholders.TIMESTAMP, timestamp),
        model_name,
    )
    writer = SummaryWriter(tb_dir) if tb_dir != "" else SummaryWriter()
    ensure_dir(file_path=tb_dir, verbose=True)

    logger.debug("Write evaluation metrics/images to tensorboard-dir: '{}'".format(tb_dir))

    img_directory_id_dict: Dict[str, int] = dict()

    for class_name, image_dict in metric_image_info_dict.items():
        for image_path, metric_image_info in image_dict.items():
            image_id, img_directory_id_dict = generate_img_id_map(
                image_path=image_path, img_directory_id_dict=img_directory_id_dict
            )

            if metric_image_info.ground_truth_annotation is not None:
                gt_count = len(
                    metric_image_info.ground_truth_annotation.get_bounding_boxes(
                        include_segmentations=True
                    )
                )
            else:
                gt_count = 0

            if metric_image_info.false_positive_annotation is not None:
                fp_count = len(
                    metric_image_info.false_positive_annotation.get_bounding_boxes(
                        include_segmentations=True
                    )
                )
            else:
                fp_count = 0

            if metric_image_info.false_negative_annotation is not None:
                fn_count = len(
                    metric_image_info.false_negative_annotation.get_bounding_boxes(
                        include_segmentations=True
                    )
                )
            else:
                fn_count = 0

            if gt_count > 0 or fp_count > 0 or fn_count > 0:
                writer.add_images(
                    tag=f"metric_image_info_{class_name}/"
                    f"{os.path.basename(image_path)}_"
                    f"ID_{image_id}_"
                    f"GT_{gt_count}_"
                    f"FP_{fp_count}_"
                    f"FN_{fn_count}",
                    img_tensor=__create_tf_image(
                        image_path=image_path,
                        metric_image_info=metric_image_info,
                        false_positive_image_size=tb_logging_config.false_positive_image_size,
                    ),
                    global_step=0,
                    dataformats="HWC",
                )

        writer.close()


def output_evaluation_results(
    model_metrics: ODModelEvaluationMetrics,
    iou_thresholds: List[float],
    score_threshold: Optional[float] = None,
    nms_threshold: Optional[float] = None,
    tensorboard_logging: Optional[TensorboardLoggingConfig] = None,
) -> None:
    """
    Generate logging output.
    - python logging
    - log object detection metrics to mlflow
    - visualize information about false positives and false negatives by logging
      images to tensorboard

    Args:
        model_metrics: An ODModelEvaluationMetrics object storing the information of an
                       evaluation computation
        iou_thresholds: Iou thresholds for which to output_evaluation_results. They will be
                        used to extract the information out of the model_metrics
        score_threshold: Optionally hand over the score threshold, indicating the score
                         threshold that has been used to filter predicted bounding boxes
        nms_threshold: Optionally hand over the nms threshold, indicating the nms
                       threshold that has been used to filter predicted bounding boxes
        tensorboard_logging: A TensorboardLoggingConfig object defining the behavior for logging
                             the visualizations of false-positives and false-negatives to
                             tensorboard

    Returns:
        None
    """

    for iou_threshold in iou_thresholds:
        metric_table = generate_metric_table(
            metrics_dict=model_metrics.metrics_dict,
            iou_threshold=iou_threshold,
        )

        score_string = f"{score_threshold:.2f}" if score_threshold is not None else "NONE"
        nms_string = f"{nms_threshold:.2f}" if nms_threshold is not None else "NONE"

        logger.info(
            "\n\n"
            " Evaluation result for "
            "model '%s', "
            "IOU= %.2f, "
            "SCORE= %s, "
            "NMS= %s: \n"
            "%s"
            "\n\n",
            model_metrics.model_specifier,
            iou_threshold,
            score_string,
            nms_string,
            metric_table.table,
        )

        log_od_metrics_to_mlflow(
            model_specifier=model_metrics.model_specifier,
            metrics_dict=model_metrics.metrics_dict,
            score_threshold=score_threshold,
            nms_threshold=nms_threshold,
            iou_threshold=iou_threshold,
        )

    log_false_positive_info(
        model_specifier=model_metrics.model_specifier,
        metric_image_info_dict=model_metrics.metrics_image_info_dict,
    )

    if tensorboard_logging is not None:
        log_false_positive_info_to_tb(
            model_name=model_metrics.model_specifier,
            metric_image_info_dict=model_metrics.metrics_image_info_dict,
            tb_logging_config=tensorboard_logging,
        )



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for creating a look-up model based on existing annotations"""
import hashlib
import logging
import os
import typing
from abc import ABC
from typing import Dict, Generic, List, Optional, Tuple, TypeVar, Union

import cv2

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.segmentation import Segmentation
from mlcvzoo_base.api.data.types import ImageType
from mlcvzoo_base.api.interfaces import Classifiable
from mlcvzoo_base.api.model import (
    ClassificationModel,
    Model,
    ObjectDetectionModel,
    SegmentationModel,
)
from mlcvzoo_base.configuration.utils import (
    create_configuration as create_basis_configuration,
)
from mlcvzoo_base.data_preparation.annotation_handler import AnnotationHandler
from mlcvzoo_base.models.read_from_file.configuration import ReadFromFileConfig

logger = logging.getLogger(__name__)


ModelType = TypeVar(
    "ModelType",
    ClassificationModel[ReadFromFileConfig, Union[str, ImageType]],
    ObjectDetectionModel[ReadFromFileConfig, Union[str, ImageType]],
    SegmentationModel[ReadFromFileConfig, Union[str, ImageType]],
)


class ReadFromFileModel(
    Model,  # type: ignore
    Classifiable,
    ABC,
    Generic[ModelType],
):
    """
    Simple Model which can be used as fast Online detector.
    It takes an "AnnotationHandlerConfig" and parses all annotations
    based on this config into a datastructure. At prediction step it
    simply looks up the annotation based on the image-path information.

    NOTE: The ReadFromFileModel is only a 'Model'. In order to become a model of a
          dedicated type like ClassificationModel or ObjectDetectionModel, the
          Subclasses of a ReadFromFileModel have to inherit not only from the
          ReadFromFileModel but also from the dedicated type itself.

    NOTE: The constructor of the super class Model will not be called since
          the ReadFromFileModel is an abstract super class and therefore is
          not intended to be instantiated. But make sure to call the Model
          constructor in one of the implementing subclasses.
    """

    def __init__(
        self,
        from_yaml: Optional[str] = None,
        configuration: Optional[ReadFromFileConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ):
        self.configuration: ReadFromFileConfig = ReadFromFileModel.create_configuration(
            from_yaml=from_yaml,
            configuration=configuration,
            string_replacement_map=string_replacement_map,
        )
        Model.__init__(
            self,
            configuration=self.configuration,
            init_for_inference=True,
        )
        Classifiable.__init__(
            self,
            mapper=AnnotationClassMapper(
                class_mapping=self.configuration.class_mapping,
                reduction_mapping=self.configuration.reduction_class_mapping,
            ),
        )

        if self.configuration.annotation_handler_config.class_mapping is not None:
            logger.warning(
                "Explicit annotation_handler_config.class_mapping is not "
                "supported in a models context anymore!"
            )

        if self.configuration.annotation_handler_config.reduction_class_mapping is not None:
            logger.warning(
                "Explicit annotation_handler_config.reduction_class_mapping is not "
                "supported in a models context anymore!"
            )

        self.annotation_handler = AnnotationHandler(
            mapper=self.mapper,
            configuration=self.configuration.annotation_handler_config,
        )

        self.annotations_dict: Dict[str, BaseAnnotation] = self.initialize_annotations_dict(
            annotations=self.annotation_handler.parse_inference_annotations()
        )

    @staticmethod
    def create_configuration(
        from_yaml: Optional[str] = None,
        configuration: Optional[ReadFromFileConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> ReadFromFileConfig:
        return typing.cast(
            ReadFromFileConfig,
            create_basis_configuration(
                configuration_class=ReadFromFileConfig,
                from_yaml=from_yaml,
                input_configuration=configuration,
                string_replacement_map=string_replacement_map,
            ),
        )

    @property
    def num_classes(self) -> int:
        return self.annotation_handler.mapper.num_classes

    def get_classes_id_dict(self) -> Dict[int, str]:
        return self.annotation_handler.mapper.annotation_class_id_to_model_class_name_map

    @staticmethod
    def create_image_hash(data_item: Union[str, ImageType]) -> str:
        if isinstance(data_item, str):
            image = cv2.imread(data_item)
        else:
            image = data_item

        return hashlib.md5(image).hexdigest()

    def initialize_annotations_dict(
        self, annotations: List[BaseAnnotation]
    ) -> Dict[str, BaseAnnotation]:
        """
        Create a hashed annotation dictionary from a given list of annotations

        Args:
            annotations: list of annotations

        Returns:
            a hashed dict with annotations
        """
        annotations_dict: Dict[str, BaseAnnotation] = {}

        for annotation in annotations:
            if annotation.image_path in annotations_dict:
                logger.warning(
                    "Duplicate annotation data for image '%s'.\n"
                    "Check your annotation-handler configuration!",
                    annotation.image_path,
                )
            if self.configuration.use_image_name_hash:
                annotations_dict[
                    os.path.basename(annotation.annotation_path).replace(".xml", "")
                ] = annotation
            else:
                annotations_dict[annotation.image_path] = annotation
                annotations_dict[self.create_image_hash(annotation.image_path)] = annotation

        return annotations_dict


class ReadFromFileClassificationModel(
    ClassificationModel[ReadFromFileConfig, Union[str, ImageType]],
    ReadFromFileModel[ClassificationModel[ReadFromFileConfig, Union[str, ImageType]]],
):
    def __init__(
        self,
        from_yaml: Optional[str] = None,
        configuration: Optional[ReadFromFileConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
        init_for_inference: bool = False,
    ):
        ReadFromFileModel.__init__(
            self,
            from_yaml=from_yaml,
            configuration=configuration,
            string_replacement_map=string_replacement_map,
        )
        ClassificationModel.__init__(
            self,
            configuration=self.configuration,
            mapper=self.mapper,
            init_for_inference=True,
        )

    def predict(
        self, data_item: Union[str, ImageType]
    ) -> Tuple[Union[str, ImageType], List[Classification]]:
        if isinstance(data_item, str):
            data_hash = data_item
        else:
            data_hash = self.create_image_hash(data_item)

        if data_hash in self.annotations_dict:
            classifications = self.annotations_dict[data_hash].classifications
            return data_item, classifications

        raise ValueError("data_item='%s' not in lookup dict of the ReadFromFileModel" % data_item)


class ReadFromFileObjectDetectionModel(
    ObjectDetectionModel[ReadFromFileConfig, Union[str, ImageType]],
    ReadFromFileModel[ObjectDetectionModel[ReadFromFileConfig, Union[str, ImageType]]],
):
    def __init__(
        self,
        from_yaml: Optional[str] = None,
        configuration: Optional[ReadFromFileConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
        init_for_inference: bool = False,
    ):
        ReadFromFileModel.__init__(
            self,
            from_yaml=from_yaml,
            configuration=configuration,
            string_replacement_map=string_replacement_map,
        )
        ObjectDetectionModel.__init__(
            self,
            configuration=self.configuration,
            mapper=self.mapper,
            init_for_inference=True,
        )

    def predict(
        self, data_item: Union[str, ImageType]
    ) -> Tuple[Union[str, ImageType], List[BoundingBox]]:
        include_segmentations: bool = self.configuration.include_segmentations

        if isinstance(data_item, str):
            data_hash = data_item
        else:
            data_hash = self.create_image_hash(data_item)

        if data_hash in self.annotations_dict:
            bounding_boxes = self.annotations_dict[data_hash].get_bounding_boxes(
                include_segmentations=include_segmentations
            )
            return data_item, bounding_boxes

        raise ValueError("data_item='%s' not in lookup dict of the ReadFromFileModel" % data_item)


class ReadFromFileSegmentationModel(
    SegmentationModel[ReadFromFileConfig, Union[str, ImageType]],
    ReadFromFileModel[SegmentationModel[ReadFromFileConfig, Union[str, ImageType]]],
):
    def __init__(
        self,
        from_yaml: Optional[str] = None,
        configuration: Optional[ReadFromFileConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
        init_for_inference: bool = False,
    ):
        ReadFromFileModel.__init__(
            self,
            from_yaml=from_yaml,
            configuration=configuration,
            string_replacement_map=string_replacement_map,
        )
        SegmentationModel.__init__(
            self,
            configuration=self.configuration,
            mapper=self.mapper,
            init_for_inference=True,
        )

    def predict(
        self, data_item: Union[str, ImageType]
    ) -> Tuple[Union[str, ImageType], List[Segmentation]]:
        if isinstance(data_item, str):
            data_hash = data_item
        else:
            data_hash = self.create_image_hash(data_item)

        if data_hash in self.annotations_dict:
            segmentations = self.annotations_dict[data_hash].segmentations

            return data_item, segmentations

        raise ValueError("data_item='%s' not in lookup dict of the ReadFromFileModel" % data_item)



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for different utility operations regarding annotations in lists"""
import logging
import random
from typing import List, Optional, Tuple

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.dataset_info import BaseDatasetInfo
from mlcvzoo_base.utils.file_utils import ensure_dir

logger = logging.getLogger(__name__)


def create_list_split(
    input_file_list: List[BaseAnnotation],
    split_size: float,
    random_state: Optional[int] = None,
) -> Tuple[List[BaseAnnotation], List[BaseAnnotation]]:
    """
    Splits the given input_file_list into training and evaluation lists.

    Args:
        input_file_list: List of BaseAnnotation objects
        split_size: float, size of split between 0 and 1
        random_state: Optional int, defining the random state for reproducible splits

    Returns: two lists of BaseAnnotation objects for training and evaluation

    """
    # create the split
    if len(input_file_list) < 2:
        raise ValueError("Can't split a list with less than 2 elements")

    file_list = input_file_list.copy()
    rng = random.Random()
    if random_state is not None:
        rng.seed(random_state)

    rng.shuffle(file_list)

    # lists for storing the path of the images
    max_index_eval = max(1, int(len(file_list) * split_size))
    eval_list = file_list[:max_index_eval]
    train_list = file_list[max_index_eval:]

    logger.info(
        "Split: train_size={} ({} images), eval_size={} ({} images)".format(
            str(1.0 - split_size),
            str(len(train_list)),
            str(split_size),
            str(len(eval_list)),
        )
    )

    return train_list, eval_list


def create_cross_val_list_splits(
    input_file_list: List[BaseAnnotation], number_splits: int
) -> List[List[BaseAnnotation]]:
    # TODO: implement or remove

    raise ValueError("create_cross_val_list_splits is not implemented for now!")


def create_annotation_file_from_list(
    csv_entry_list: List[str], dataset_info: BaseDatasetInfo, output_file_path: str
) -> BaseDatasetInfo:
    """
    Creates an annotation file from csv annotations

    Args:
        csv_entry_list: List of strings (csv entries)
        dataset_info: BaseDatasetInfo, basic information about a dataset
        output_file_path: string, defines the path where the file should be stored

    Returns: BaseDatasetInfo, the updated information

    """

    dataset_info.base_file = output_file_path

    ensure_dir(file_path=output_file_path, verbose=True)

    logger.info("Write annotations to file: %s", output_file_path)
    with open(output_file_path, "w", encoding="'utf-8") as file:
        file.writelines(csv_entry_list)

    return dataset_info


def generate_tags_based_on_bounding_boxes(
    annotations: List[BaseAnnotation],
) -> List[BaseAnnotation]:
    """

    Args:
        annotations: List of BaseAnnotations from which tags are derived

    Returns:
        List of BaseAnnotations enriched with Classifications

    """
    class_identifier = ClassIdentifier(class_id=-1, class_name="")
    default_bounding_box = BoundingBox(
        box=Box(0, 0, 0, 0),
        class_identifier=class_identifier,
        model_class_identifier=class_identifier,
        score=0.0,
        difficult=False,
        occluded=False,
        content="",
    )

    for annotation in annotations:
        # TODO:
        #  get biggest bounding box out of
        #  annotation.get_bounding_boxes(include_segmentations=True)

        max_bounding_box: BoundingBox = default_bounding_box
        max_area = 0
        for bounding_box in annotation.get_bounding_boxes(include_segmentations=True):
            width = bounding_box.box.xmax - bounding_box.box.xmin
            height = bounding_box.box.ymax - bounding_box.box.ymin
            area = width * height
            if area > max_area:
                max_bounding_box = bounding_box

        assert not max_bounding_box.__eq__(default_bounding_box)
        classification: Classification = Classification(
            class_identifier=max_bounding_box.class_identifier,
            model_class_identifier=max_bounding_box.model_class_identifier,
            score=1.0,
        )

        annotation.classifications.append(classification)

    return annotations



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for parsing information from yaml in python accessible attributes for different
configuration classes and also for the ModelRegistry class.
"""

from inspect import getfullargspec
from typing import Any, Dict, List, Optional, Type, Union, cast, get_args

import related
from attr import define
from config_builder import BaseConfigClass

from mlcvzoo_base.configuration.utils import str2bool

INIT_FOR_INFERENCE_PARAMETER = "init_for_inference"


@define
class ModelConfig(BaseConfigClass):
    __related_strict__ = True

    class_type: str = related.StringField()
    constructor_parameters: Dict[
        str, Union[int, bool, str, float, List[Any], Dict[str, Any]]
    ] = related.ChildField(
        cls=dict,
    )

    def update_class_type(self, args_dict: Dict[str, Optional[str]]) -> None:
        if "class_type" in args_dict and args_dict["class_type"] is not None:
            self.class_type = args_dict["class_type"]

    def update_constructor_parameters(
        self, args_dict: Dict[str, Optional[str]], model_type: Type[Any]
    ) -> None:
        model_arg_spec = getfullargspec(model_type.__init__)

        if (
            "constructor_parameters" in args_dict
            and args_dict["constructor_parameters"] is not None
        ):
            constructor_parameters = {}
            for constructor_parameters_dict in args_dict["constructor_parameters"]:
                constructor_parameters.update(cast(Dict[str, Any], constructor_parameters_dict))

            for (
                constructor_parameter_key,
                constructor_parameter_value,
            ) in constructor_parameters.items():
                if constructor_parameter_key in model_arg_spec.annotations.keys():
                    type_arg = model_arg_spec.annotations[constructor_parameter_key]

                    # In case of a "typing.Optional" parameter type, the casting_type
                    # has to be determined by utilizing "typing.get_args"
                    if len(get_args(type_arg)) == 0:
                        casting_type = type_arg
                    else:
                        casting_type = get_args(type_arg)[0]

                    # bool types have to be handled by str2bool
                    if casting_type is bool:
                        self.constructor_parameters[constructor_parameter_key] = str2bool(
                            constructor_parameter_value
                        )
                    else:
                        self.constructor_parameters[constructor_parameter_key] = casting_type(
                            constructor_parameter_value
                        )

    def is_inference(self) -> bool:
        return INIT_FOR_INFERENCE_PARAMETER in self.constructor_parameters

    def set_inference(self, inference: bool) -> None:
        self.constructor_parameters[INIT_FOR_INFERENCE_PARAMETER] = inference



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Data class for storing information that is needed to classify an object instance"""

from __future__ import annotations

from typing import Any, Dict


class ClassIdentifier:
    __delimiter: str = "_"

    def __init__(
        self,
        class_id: int,
        class_name: str,
    ):
        # id of the 'class', respectively the id of the object type
        self.__class_id: int = class_id

        # name of the 'class', respectively the name of the object type
        self.__class_name: str = class_name

    def __eq__(self, other: ClassIdentifier) -> bool:  # type: ignore
        return self.class_id == other.class_id or self.class_name == other.class_name

    def __hash__(self) -> int:
        return hash((self.__class_id, self.__class_name))

    def __repr__(self) -> str:
        return f"{self.class_id}{ClassIdentifier.__delimiter}{self.class_name}"

    def to_dict(self) -> Dict[str, Any]:
        return {"class_id": self.class_id, "class_name": self.class_name}

    def to_json(self) -> Any:
        return self.to_dict()

    @property
    def class_id(self) -> int:
        return self.__class_id

    @property
    def class_name(self) -> str:
        return self.__class_name

    @staticmethod
    def from_str(class_identifier_str: str) -> ClassIdentifier:
        """
        Build a ClassIdentifier object from the given string

        Args:
            class_identifier_str: The string to build the ClassIdentifier from

        Returns:
            The built ClassIdentifier object
        """
        split_position = class_identifier_str.find(ClassIdentifier.__delimiter)

        # The class_identifier_str is expected to be in the format:
        #    CLASSID{ClassIdentifier.__delimiter}CLASSNAME
        # Therefore, besides providing the delimiter, we need at
        # least one character for the class-id and one character
        # following the delimiter for the class-name
        if (
            split_position == -1
            or split_position == 0
            or len(class_identifier_str) <= split_position + len(ClassIdentifier.__delimiter)
        ):
            raise ValueError(
                f"Could not build ClassIdentifier from '{class_identifier_str}'. "
                f"Please provide it in the format: "
                f"'CLASSID{ClassIdentifier.__delimiter}CLASSNAME'"
            )

        return ClassIdentifier(
            class_id=int(class_identifier_str[0:split_position]),
            class_name=class_identifier_str[split_position + 1 :],
        )



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for parsing information from yaml in python accessible attributes about the program
environment for different configuration classes.
"""
import related
from attr import define
from config_builder import BaseConfigClass

STRING_REPLACEMENT_MAP_KEY = "string_replacement_map"


@define
class ReplacementConfig(BaseConfigClass):
    """
    Class for parsing general information about the environment placeholder mapping
    """

    CONFIG_ROOT_DIR_DIR_KEY = "CONFIG_ROOT_DIR"
    BASELINE_MODEL_DIR_KEY = "BASELINE_MODEL_DIR"
    INFERENCE_MODEL_DIR_KEY = "INFERENCE_MODEL_DIR"
    TRAINING_MODEL_DIR_KEY = "TRAINING_MODEL_DIR"
    MMDETECTION_DIR_KEY = "MMDETECTION_DIR"
    MMOCR_DIR_KEY = "MMOCR_DIR"
    MEDIA_DIR_KEY = "MEDIA_DIR"
    ANNOTATION_DIR_KEY = "ANNOTATION_DIR"
    PROJECT_ROOT_DIR_KEY = "PROJECT_ROOT_DIR"
    DARKNET_DIR_KEY = "DARKNET_DIR"
    YOLOX_DIR_KEY = "YOLOX_DIR"
    MLCVZOO_DIR_KEY = "MLCVZOO_DIR"

    DEFAULT_MAP = {
        CONFIG_ROOT_DIR_DIR_KEY: "",
        BASELINE_MODEL_DIR_KEY: "",
        INFERENCE_MODEL_DIR_KEY: "",
        TRAINING_MODEL_DIR_KEY: "",
        MMDETECTION_DIR_KEY: "",
        MMOCR_DIR_KEY: "",
        MEDIA_DIR_KEY: "",
        ANNOTATION_DIR_KEY: "",
        PROJECT_ROOT_DIR_KEY: "",
        DARKNET_DIR_KEY: "",
        YOLOX_DIR_KEY: "",
        MLCVZOO_DIR_KEY: "",
    }

    # Root Directory of the current repository
    PROJECT_ROOT_DIR: str = related.StringField(default=DEFAULT_MAP[PROJECT_ROOT_DIR_KEY])

    # Root directory of all configuration files
    CONFIG_ROOT_DIR: str = related.StringField(default=DEFAULT_MAP[CONFIG_ROOT_DIR_DIR_KEY])

    # Directory where the project mlcvzoo has be cloned to
    MLCVZOO_DIR: str = related.StringField(default=DEFAULT_MAP[MLCVZOO_DIR_KEY])

    # ==================================================
    # Placeholder for checkpoints directories

    # Main directory for all checkpoint files of original models
    # (e.g. models trained on the coco dataset)
    BASELINE_MODEL_DIR: str = related.StringField(default=DEFAULT_MAP[BASELINE_MODEL_DIR_KEY])

    # Main directory for storing inference model/checkpoint files only. When you just want to
    # use a certain device for inference, you don't want to synchronize all the training
    # checkpoints, but only the checkpoints that are relevant for the inference.
    INFERENCE_MODEL_DIR: str = related.StringField(default=DEFAULT_MAP[INFERENCE_MODEL_DIR_KEY])

    # Main directory for all shared training model/checkpoint files
    TRAINING_MODEL_DIR: str = related.StringField(default=DEFAULT_MAP[TRAINING_MODEL_DIR_KEY])

    # ==================================================
    # Placeholder for external repositories

    # Directory where the project mmdetection-fork has be cloned to
    MMDETECTION_DIR: str = related.StringField(default=DEFAULT_MAP[MMDETECTION_DIR_KEY])

    # Directory where the project mmdetection-fork has be cloned to
    MMOCR_DIR: str = related.StringField(default=DEFAULT_MAP[MMOCR_DIR_KEY])

    # Directory where the project AlexeyAB/darknet has be cloned and compiled to
    DARKNET_DIR: str = related.StringField(default=DEFAULT_MAP[DARKNET_DIR_KEY])

    # Directory where the project yolox has be cloned to
    YOLOX_DIR: str = related.StringField(default=DEFAULT_MAP[YOLOX_DIR_KEY])

    # ==================================================
    # Placeholder for data directories

    # Directory where the central media data is stored
    MEDIA_DIR: str = related.StringField(default=DEFAULT_MAP[MEDIA_DIR_KEY])

    # Directory where the central annotation data is stored
    ANNOTATION_DIR: str = related.StringField(default=DEFAULT_MAP[ANNOTATION_DIR_KEY])



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Definition of interfaces / abstract classes that define additional features of mlcvzoo models.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, Generic, Optional, TypeVar

from mlcvzoo_base.api.configuration import InferenceConfig
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper

NetType = TypeVar("NetType")
NetConfigurationType = TypeVar("NetConfigurationType", bound=InferenceConfig)


@dataclass
class Perception:
    """
    Base class for representing any data object which is consumed by
    models of the MLCVZoo. This includes ground-truth data as well as
    predicted objects. Dataclasses should inherit from this class in
    order to be used in a predict(...) method of the model class,
    or subclasses of the model class e.g. for any object detector,
    character classifier, OCR, etc.
    """


class Classifiable(ABC):
    """
    A model that inherits from the Classifiable interface states that its prediction
    output is representing some kind of class instance. A class in this context means
    real world instance like person, car, truck, cat, ... etc.
    """

    def __init__(self, mapper: AnnotationClassMapper):
        self._mapper = mapper

    @property
    def mapper(self) -> AnnotationClassMapper:
        """
        Returns:
            The mapper object associated with this classifiable model
        """
        return self._mapper

    @property
    def num_classes(self) -> int:
        """
        Returns:
            The number of classes for which this model is build.
        """
        raise NotImplementedError("Must be implemented by sub-class: num_classes(...).")

    def get_classes_id_dict(self) -> Dict[int, str]:
        """
        The prediction output of classifiable models contains representations of classes
        as IDs. These IDs are associated with real world instances like car, truck, person etc.
        Therefore, each classifiable model has to provide a dictionary that indicates how the
        class IDs are mapped to class names.

        Returns:
            The dictionary that indicates the mapping of class IDs to class names
        """
        raise NotImplementedError("Must be implemented by sub-class: get_classes_id_dict(...).")


class Trainable(ABC):
    """
    The interface represents the super-class for any trainable model.
    A trainable model must not necessarily be a model based on
    an neural network. A trainable model could also be an algorithm like
    any reinforcement algorithm for example.
    """

    @abstractmethod
    def train(self) -> None:
        """
        The method trains the model.

        Subclasses must implement the training, including proper use of the
        model configuration and parameters.

        Returns:
            None
        """
        raise NotImplementedError("Must be implemented by sub-class: train(...).")

    @abstractmethod
    def get_training_output_dir(self) -> str:
        """
        Returns:
            The directory where a trainable model is storing its training output artefacts
        """
        raise NotImplementedError(
            "Must be implemented by sub-class: get_training_output_dir(...)."
        )


class NetBased(ABC, Generic[NetType, NetConfigurationType]):
    """
    The interface represents the super-class for any net based model. The term "net"
    in this context is associated with any kind of neural network.
    """

    def __init__(self, net: Optional[NetType]):
        # A net attribute is not always necessary, e.g. there are trainable models,
        # that don't require a net attribute during training
        self.net: Optional[NetType] = net

    @property
    def inference_config(self) -> Optional[NetConfigurationType]:
        """
        Returns:
           The inference config of this NetBased instance
        """
        return None

    def get_net(self) -> Optional[NetType]:
        """
        Returns:
            An instance of the net used in this model.
        """
        return self.net

    @abstractmethod
    def store(self, checkpoint_path: str) -> None:
        """
        The method stores the weights of the net in the given checkpoint path
        """
        raise NotImplementedError("Must be implemented by sub-class: store(...)")

    @abstractmethod
    def restore(self, checkpoint_path: str) -> None:
        """
        The method restores the weights of the net from the given checkpoint path

        """
        raise NotImplementedError("Must be implemented by sub-class: restore(...)")

    def get_checkpoint_filename_suffix(self) -> str:
        """
        Returns:
            The suffix of the associated checkpoint files like .pth (pytorch) or .h5 (tensorflow)
        """
        return ""



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for creating machine learning models in the MLCVZoo"""

from abc import ABC, abstractmethod
from typing import Dict, Generic, List, Optional, Tuple, TypeVar

from mlcvzoo_base.api.configuration import ModelConfiguration
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.ocr_perception import OCRPerception
from mlcvzoo_base.api.data.segmentation import PolygonType, Segmentation
from mlcvzoo_base.api.interfaces import Classifiable, Perception
from mlcvzoo_base.configuration.structs import ObjectDetectionBBoxFormats

DataType = TypeVar("DataType")
PredictionType = TypeVar("PredictionType", bound=Perception)
ConfigurationType = TypeVar("ConfigurationType", bound=ModelConfiguration)


class Model(ABC, Generic[PredictionType, ConfigurationType, DataType]):
    """
    A model to detect objects, like things from an image or
    characters from a license plate or book.
    """

    def __init__(
        self,
        configuration: ConfigurationType,
        init_for_inference: bool,
    ):
        """
        Constructor creates a new model instance.
        This must be called by subclass constructors to define a unique model name.
        """
        self.configuration: ConfigurationType = configuration

        if init_for_inference:
            self._init_inference_model()
        else:
            self._init_training_model()

    @property
    def unique_name(self) -> str:
        return self.configuration.unique_name

    def _init_inference_model(self) -> None:
        # Not every model differentiates between an initialization for inference and training
        pass

    def _init_training_model(self) -> None:
        # Not every model differentiates between an initialization for inference and training
        pass

    def get_configuration(self) -> ConfigurationType:
        """
        Returns:
            The configuration for the model subclass instance.The model is responsible
            to read and create its configuration, e.g. from a file.
        """
        return self.configuration

    @staticmethod
    @abstractmethod
    def create_configuration(
        from_yaml: Optional[str] = None,
        configuration: Optional[ConfigurationType] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> ConfigurationType:
        """
        The model is responsible to provide a configuration object. This encapsulates all
        relevant parameters for the model. The configuration will either be parsed from the
        file given by the "from_yaml" parameter or the provided configuration object is simply
        passed-trough.

        NOTE: This method is meant to be integrated and used with the config-builder module.

        Args:
            from_yaml: (Optional) Parse the configuration from this file path
            configuration: (Optional) When provided this configuration object is passed-trough.
                           It is leveraged as a convenience option.
            string_replacement_map: (Optional) A dictionary that defines placeholders which can
                                    be used while parsing the file. They can be understood as
                                    variables that can be used to define configs that are valid
                                    across multiple devices.

        Returns:
            The created configuration object for this model
        """
        raise NotImplementedError("Must be implemented by sub-class: create_configuration(...).")

    @abstractmethod
    def predict(self, data_item: DataType) -> Tuple[DataType, List[PredictionType]]:
        """
        The method predicts a list of classifications from a given data item.

        Args:
            data_item: The data item as input for the model

        Returns:
            A tuple containing the input value and its predictions
        """
        raise NotImplementedError("Must be implemented by sub-class: predict(...).")

    def predict_many(
        self, data_items: List[DataType]
    ) -> List[Tuple[DataType, List[PredictionType]]]:
        """
        The method predicts a list of classifications from a given data item.

        Args:
            data_items: The data item as input for the model

        Returns:
            A tuple containing the input value and its predictions
        """
        results: List[Tuple[DataType, List[PredictionType]]] = []
        for data_item in data_items:
            results.append(self.predict(data_item=data_item))

        return results


class ClassificationModel(
    Model[Classification, ConfigurationType, DataType],
    Classifiable,
    ABC,
):
    """
    Class that declares the generic method predict(...) for all models that are
    performing the computer vision task "Image Classification".
    """

    def __init__(
        self,
        configuration: ConfigurationType,
        mapper: AnnotationClassMapper,
        init_for_inference: bool,
    ):
        Classifiable.__init__(self, mapper=mapper)
        Model.__init__(
            self,
            configuration=configuration,
            init_for_inference=init_for_inference,
        )

    @abstractmethod
    def predict(self, data_item: DataType) -> Tuple[DataType, List[Classification]]:
        raise NotImplementedError("Must be implemented by sub-class: predict(...).")


class ObjectDetectionModel(
    Model[BoundingBox, ConfigurationType, DataType],
    Classifiable,
    ABC,
):
    """
    Class that declares the generic method predict(...) for all models that are
    performing the computer vision task "Object Detection".
    """

    def __init__(
        self,
        configuration: ConfigurationType,
        mapper: AnnotationClassMapper,
        init_for_inference: bool,
    ):
        Classifiable.__init__(self, mapper=mapper)
        Model.__init__(
            self,
            configuration=configuration,
            init_for_inference=init_for_inference,
        )

    @abstractmethod
    def predict(self, data_item: DataType) -> Tuple[DataType, List[BoundingBox]]:
        raise NotImplementedError("Must be implemented by sub-class: predict(...).")

    @staticmethod
    def build_bounding_boxes(
        box_list: Tuple[int, int, int, int],
        class_identifiers: List[ClassIdentifier],
        model_class_identifier: ClassIdentifier,
        score: float = 1.0,
        box_format: str = ObjectDetectionBBoxFormats.XYXY,
        difficult: bool = False,
        occluded: bool = False,
        content: str = "",
        src_shape: Optional[Tuple[int, int]] = None,
        dst_shape: Optional[Tuple[int, int]] = None,
    ) -> List[BoundingBox]:
        """
        Creates a list of BoundingBox objects with the given specifications. The bounding box
        can be scaled by defining a src shape and a destination shape. One bounding box per
        ClassIdentifier in the given list will be created.

        Args:
            box_list: object as 4D array containing bounding box information, needed
                      to create a Box object
            class_identifiers: List of ClassIdentifier objects
            model_class_identifier: List of ClassIdentifier objects
            score: classification score of the class that is captured within the bounding box
            box_format: specify the way for parsing the box argument
            difficult: Whether the object is difficult
            occluded: Whether the object is occluded
            content: content of the boxed area, e.g. the text stated in it
            src_shape: source shape for applying a scaling of the box (height, width)
            dst_shape: destination shape for applying a scaling of the box (height, width)

        Returns:
            The created list of BoundingBoxes
        """

        box = Box.init_format_based(
            box_list=box_list,
            box_format=box_format,
            src_shape=src_shape,
            dst_shape=dst_shape,
        )

        bounding_boxes: List[BoundingBox] = []

        for class_identifier in class_identifiers:
            bounding_boxes.append(
                BoundingBox(
                    difficult=difficult,
                    occluded=occluded,
                    content=content,
                    class_identifier=class_identifier,
                    model_class_identifier=model_class_identifier,
                    score=score,
                    box=box,
                )
            )

        return bounding_boxes


class SegmentationModel(Model[Segmentation, ConfigurationType, DataType], Classifiable, ABC):
    """
    Class that declares the generic method predict(...) for all models that are
    performing the computer vision task "Instance Segmentation".
    """

    def __init__(
        self,
        configuration: ConfigurationType,
        mapper: AnnotationClassMapper,
        init_for_inference: bool,
    ):
        Classifiable.__init__(self, mapper=mapper)
        Model.__init__(
            self,
            configuration=configuration,
            init_for_inference=init_for_inference,
        )

    @abstractmethod
    def predict(self, data_item: DataType) -> Tuple[DataType, List[Segmentation]]:
        raise NotImplementedError("Must be implemented by sub-class: predict(...).")

    @staticmethod
    def build_segmentations(
        polygon: PolygonType,
        class_identifiers: List[ClassIdentifier],
        model_class_identifier: Optional[ClassIdentifier] = None,
        score: float = 1.0,
        difficult: bool = False,
        occluded: bool = False,
        content: str = "",
    ) -> List[Segmentation]:
        """
        Creates a list of Segmentation objects with the given specifications.
        One Segmentation per ClassIdentifier in the given list will be created.

        Args:
            polygon: a list of points 2D points (tuples) that form the polygon
            class_identifiers: List of ClassIdentifier objects
            model_class_identifier: List of ClassIdentifier objects
            score: classification score of the area the object boxes
            difficult: Whether the object is difficult
            occluded: Whether the object is occluded
            content:  content of the captured area, e.g. the text stated in it

        Returns:
            A Segmentation Object
        """

        box: Box = Segmentation.polygon_to_box(polygon=polygon)

        segmentations: List[Segmentation] = []

        for class_identifier in class_identifiers:
            segmentations.append(
                Segmentation(
                    class_identifier=class_identifier,
                    model_class_identifier=model_class_identifier,
                    score=score,
                    polygon=polygon,
                    difficult=difficult,
                    occluded=occluded,
                    content=content,
                    box=box,
                )
            )

        return segmentations


# TODO: Rename to TextRecognitionModel?
class OCRModel(Model[OCRPerception, ConfigurationType, DataType], ABC):
    """
    Class that declares the generic method predict(...) for all models that are
    performing the computer vision task "Text Recognition" which is a subfield of
    Optical Character Recognition (OCR).
    """

    def __init__(
        self,
        configuration: ConfigurationType,
        init_for_inference: bool,
    ):
        Model.__init__(
            self,
            configuration=configuration,
            init_for_inference=init_for_inference,
        )

    @abstractmethod
    def predict(self, data_item: DataType) -> Tuple[DataType, List[OCRPerception]]:
        raise NotImplementedError("Must be implemented by sub-class: predict(...).")



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for parsing information from yaml in python accessible attributes for different
annotation handling classes like AnnotationClassMapper and several configuration classes
where a class mapping is used.
"""

import logging
from typing import List, Optional

import related
from attr import define
from config_builder import BaseConfigClass

logger = logging.getLogger(__name__)


@define
class ClassMappingMappingConfig(BaseConfigClass):
    """Class for parsing information about a specific mapping"""

    input_class_name: str = related.StringField()
    output_class_name: str = related.StringField()


@define
class ClassMappingModelClassesConfig(BaseConfigClass):
    """Class for parsing information about classes"""

    class_name: str = related.StringField()
    class_id: int = related.IntegerField()


@define
class ClassMappingConfig(BaseConfigClass):
    """
    The ClassMappingConfig is relevant while parsing annotations files into data structures of the
    MLCVZoo. These data structures are then used to prepare the training / ground truth data
    for any model of the MLCVZoo. The class_mapping defines the number of classes a model knows
    and how the according model class IDs relate to class names (class_mapping.model_classes)
    that are parsed from the annotation files. Furthermore, the class_mapping allows to define
    a mapping of class names via the configuration attribute class_mapping.mapping. This can be
    used to fix typos that arise during the annotation process or to aggregate classes to a
    single super class, e.g. car, truck and motorcycle are aggregated as vehicle.

    EXAMPLE:

    model_classes:
      - class_name: "person"
        class_id: 0
      - class_name: "truck"
        class_id: 1
      - class_name: "car"
        class_id: 2
    """

    model_classes: List[ClassMappingModelClassesConfig] = related.SequenceField(
        ClassMappingModelClassesConfig
    )
    mapping: List[ClassMappingMappingConfig] = related.SequenceField(
        ClassMappingMappingConfig, required=False, default=[]
    )

    # class-names that are contained in "ignore_class_names" are forbidden to be
    # contained in any annotation object. This applies for all Classifications,
    # Bounding-Boxes and Segmentations, see src/mlcvzoo/api/data/annotation.py for
    # more details.
    # If a class-name of the "ignore_class_names" is parsed, the complete annotation
    # for this image will be ignored. Example use-cases are the usage of "bad-image"
    # to indicate images, that have been put in CVAT, but should be ignored for further
    # usage.
    ignore_class_names: List[str] = related.SequenceField(cls=str, default=[])

    number_model_classes: Optional[int] = related.ChildField(cls=int, required=False, default=None)
    number_classes: Optional[int] = related.ChildField(cls=int, required=False, default=None)

    def get_number_classes(self) -> int:
        """

        Returns:
            The number of model classes of this class mapping
        """
        return (
            self.number_model_classes
            if self.number_model_classes is not None
            else self.number_classes  # type: ignore
        )

    def check_values(self) -> bool:
        if self.number_classes is not None:
            logger.warning(
                "'number_classes' of config class ClassMappingConfig "
                "is deprecated and will be removed in "
                "future versions! Please use number_model_classes instead"
            )
            return True

        if self.number_model_classes is None:
            logger.error("'number_model_classes' of config class ClassMappingConfig is not set!")
            return False

        return True



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for configuring the visualization of model outputs
"""

import related
from attr import define
from config_builder import BaseConfigClass


@define
class VisualizationConfig(BaseConfigClass):
    """Class for parsing information for visualization"""

    show_image: bool = related.BooleanField(default=True)
    font_path: str = related.StringField(default="")
    output_shape: int = related.IntegerField(default=500)



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module that defines the ReductionMappingConfig that is used in the
mlcvzoo_base.data_preparation.annotation_class_mapper.AnnotationClassMapper.
"""

import logging
from typing import Dict, List, Optional

import related
from attr import define
from config_builder import BaseConfigClass

logger = logging.getLogger(__name__)


@define
class ReductionMappingMappingConfig(BaseConfigClass):
    """
    A ReductionMappingMappingConfig entry defines how model class IDs / names should
    be mapped to an output class ID and output class name. This can be used to aggregate
    or redefine model classes.
    """

    output_class_name: str = related.StringField()
    output_class_id: int = related.IntegerField()
    model_class_ids: Optional[List[int]] = related.ChildField(
        cls=list, required=False, default=None
    )
    model_class_names: Optional[List[str]] = related.ChildField(
        cls=list, required=False, default=None
    )

    @property
    def _mutual_attributes(self) -> List[str]:
        return ["model_class_ids", "model_class_names"]

    def check_values(self) -> bool:
        success: bool = True
        if self.model_class_ids is not None:
            for model_class_id in self.model_class_ids:
                if not isinstance(model_class_id, int):
                    logger.info(
                        "model_class_ids have to be integer values! model_class_id: '%s'"
                        % self.model_class_ids
                    )
                    return False

        if self.model_class_names is not None:
            for model_class_name in self.model_class_names:
                if not isinstance(model_class_name, str):
                    logger.info(
                        "model_class_names have to be string values! model_class_names: '%s'"
                        % self.model_class_names
                    )
                    return False

        return success


@define
class ReductionMappingConfig(BaseConfigClass):
    """
    After a model has been trained, it will generate predictions that only include the class IDs
    or class names, it has been trained on. In order to be able to aggregate or adapt the relation
    between class IDs and class names the reduction_mapping can be used.
    """

    mapping: List[ReductionMappingMappingConfig] = related.SequenceField(
        ReductionMappingMappingConfig
    )



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

""" Class for Bounding Box information"""

from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional, Tuple

import cv2

from mlcvzoo_base.api.data.types import ImageType
from mlcvzoo_base.configuration.structs import ObjectDetectionBBoxFormats

logger = logging.getLogger(__name__)


class Box:
    """
    Class for storing bounding box information.

    Box on an Image:
    |-----------------------|
    |(xmin, ymin)           |
    |                       |
    |                       |
    |                       |
    |                       |
    |                       |
    |           (xmax, ymax)|
    |-----------------------|
    """

    def __init__(self, xmin: int, ymin: int, xmax: int, ymax: int) -> None:
        if xmin >= xmax:
            raise ValueError(f"xmin={xmin} has to be < xmax={xmax}")

        if ymin >= ymax:
            raise ValueError(f"ymin={ymin} has to be < ymax={ymax}")

        # top left x coordinate
        self.__xmin: int = xmin
        # top left y coordinate
        self.__ymin: int = ymin
        # lover right x coordinate
        self.__xmax: int = xmax
        # lower right y coordinate
        self.__ymax: int = ymax

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Box):
            return (
                self.xmin == other.xmin
                and self.ymin == other.ymin
                and self.xmax == other.xmax
                and self.ymax == other.ymax
            )

        return False

    def __repr__(self):  # type: ignore
        return (
            f"Box("
            f"xmin={self.__xmin}, ymin={self.__ymin}, xmax={self.__xmax}, ymax={self.__ymax}"
            f")"
        )

    def to_dict(self) -> Dict[str, int]:
        return {
            "xmin": self.xmin,
            "ymin": self.ymin,
            "xmax": self.xmax,
            "ymax": self.ymax,
        }

    def to_json(self) -> Any:
        return self.to_dict()

    @property
    def xmin(self) -> int:
        return self.__xmin

    @property
    def ymin(self) -> int:
        return self.__ymin

    @property
    def xmax(self) -> int:
        return self.__xmax

    @property
    def ymax(self) -> int:
        return self.__ymax

    @property
    def width(self) -> int:
        return self.__xmax - self.__xmin

    @property
    def height(self) -> int:
        return self.__ymax - self.__ymin

    @staticmethod
    def init_format_based(
        box_format: str,
        box_list: Tuple[int, int, int, int],
        src_shape: Optional[Tuple[int, int]] = None,
        dst_shape: Optional[Tuple[int, int]] = None,
    ) -> Box:
        """
        Additional Constructor

        Args:
            box_format: specify the way for parsing the box argument
            box_list: object as 4D array containing bounding box information
            src_shape: shape of the original image as tuple (height, width)
            dst_shape: desired shape for creating the bounding boxes as tuple (height, width)

        Returns:
            A Box object
        """

        xmin = box_list[0]
        if xmin < 0:
            logger.warning("xmin=%d coordinate < 0, clamp xmin to 0" % xmin)
            xmin = max(0, xmin)

        ymin = box_list[1]
        if ymin < 0:
            logger.warning("ymin=%d coordinate < 0, clamp ymin to 0" % ymin)
            ymin = max(0, ymin)

        if box_format == ObjectDetectionBBoxFormats.XYWH:
            width = box_list[2]
            if width < 0:
                raise ValueError("Can not build a box with negative width")

            height = box_list[3]
            if height < 0:
                raise ValueError("Can not build a box with negative height")

            base_box = Box(
                xmin=int(xmin),
                ymin=int(ymin),
                xmax=int(xmin + width),
                ymax=int(ymin + height),
            )
        elif box_format == ObjectDetectionBBoxFormats.XYXY:
            base_box = Box(
                xmin=int(xmin),
                ymin=int(ymin),
                xmax=int(box_list[2]),
                ymax=int(box_list[3]),
            )
        else:
            valid_formats = ObjectDetectionBBoxFormats.get_values_as_list(
                class_type=ObjectDetectionBBoxFormats
            )
            raise ValueError(
                f"Format {box_format} is not supported. " f"Please provide any of {valid_formats}"
            )

        if src_shape is not None:
            base_box.clamp(shape=src_shape)

        if src_shape is not None and dst_shape is not None:
            base_box.scale(src_shape=src_shape, dst_shape=dst_shape)

        return base_box

    def to_list(self, dst_type: Any = int) -> List[Any]:
        """
        Args:
            dst_type: destination type to transform the data to

        Returns:
            List of transformed polygons
        """

        return [
            dst_type(self.xmin),
            dst_type(self.ymin),
            dst_type(self.xmax),
            dst_type(self.ymax),
        ]

    def clamp(self, shape: Tuple[int, int]) -> None:
        """
        Clamps the bounding-box based on the given shape

        Args:
            shape: The shape to define the min and max coordinates
                   for the clamping, format in (y, x)

        Returns:
            None
        """
        self.__xmin = int(max(0, self.__xmin))
        self.__ymin = int(max(0, self.__ymin))

        self.__ymax = int(min(shape[0] - 1, self.__ymax))
        self.__xmax = int(min(shape[1] - 1, self.__xmax))

    def scale(self, src_shape: Tuple[int, int], dst_shape: Tuple[int, int]) -> None:
        """
        Scale the Box according to the given shapes of the source and destination image

        Args:
            src_shape: shape of the original image as tuple (height, width)
            dst_shape: desired shape for creating the bounding boxes as tuple (height, width)

        Returns:
            None
        """
        if src_shape[0] < 0 or src_shape[1] < 0 or src_shape[0] > src_shape[1]:
            raise ValueError("Invalid source shape %s: ", src_shape)

        if dst_shape[0] < 0 or dst_shape[1] < 0 or dst_shape[0] > dst_shape[1]:
            raise ValueError("Invalid destination shape %s: ", dst_shape)

        height_scale_factor = dst_shape[0] / src_shape[0]
        width_scale_factor = dst_shape[1] / src_shape[1]

        self.__xmin = int(self.xmin * width_scale_factor)
        self.__xmax = int(self.xmax * width_scale_factor)

        self.__ymin = int(self.ymin * height_scale_factor)
        self.__ymax = int(self.ymax * height_scale_factor)

    def as_array(self) -> List[int]:
        """
        Transforms the Box object to a list of coordinates

        Returns:
            A List of coordinates in the order [xmin, ymin, xmax, ymax]
        """
        return [self.xmin, self.ymin, self.xmax, self.ymax]

    def center(self) -> Tuple[float, float]:
        """
        Calculates the center coordinates of the Box

        Returns:
            A Tuple as the coordinates of the center
        """
        return (
            self.xmin + (self.xmax - self.xmin) * 0.5,
            self.ymin + (self.ymax - self.ymin) * 0.5,
        )

    def translation(self, x: int, y: int) -> None:
        """
        Shifts Box for x and y pixels in x and y direction respectively

        Args:
            x: int value for shift in x direction
            y: int value for shift in y direction

        Returns:
            None
        """
        self.__xmin += x
        self.__xmax += x
        self.__ymin += y
        self.__ymax += y

    def new_center(self, x: int, y: int) -> None:
        """
        Shifts the Box based on a new center coordinate. Scale of Box is kept.

        Args:
            x: int value, x coordinate of the new center
            y: int value, y coordinate of the new center

        Returns:
            None
        """
        width = self.xmax - self.xmin
        height = self.ymax - self.ymin

        self.__xmin = round(x - width / 2)
        self.__xmax = round(x + width / 2)
        self.__ymin = round(y - height / 2)
        self.__ymax = round(y + height / 2)

    def crop_img(
        self, frame: ImageType, margin_x: float = 0.0, margin_y: float = 0.0
    ) -> Optional[ImageType]:
        """
        Create a crop of the given frame based on the information of this box object
        and the given margins. The margin are used as scale factors based on the
        width (x direction) and height (y-direction)

        Args:
            frame: The frame to crop from
            margin_x: The margin around the box in x direction
            margin_y: The margin around the box in y direction

        Returns:
            The cropped image (if it could be computed)
        """

        if frame is None:
            return None

        xmin = int(self.xmin - int(round(margin_x * (self.xmax - self.xmin))))
        xmax = int(self.xmax + int(round(margin_x * (self.xmax - self.xmin))))
        ymin = int(self.ymin - int(round(margin_y * (self.ymax - self.ymin))))
        ymax = int(self.ymax + int(round(margin_y * (self.ymax - self.ymin))))

        margin_box = Box(xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax)
        margin_box.clamp(shape=(frame.shape[0], frame.shape[1]))

        cropped_image = frame[margin_box.ymin : margin_box.ymax, margin_box.xmin : margin_box.xmax]

        return cropped_image

    def color_hist(
        self,
        frame: ImageType,
        margin_x: float,
        margin_y: float,
    ) -> Optional[ImageType]:
        """
        Calculate the color history the for the Box. First a crop
        of the given image is created based on the box information
        of this object and the given margins. The margin are used
        as scale factors based on the width (x direction) and height
        (y-direction)Afterwards the histogram is computed for this crop.

        Args:
            frame: The frame to crop from
            margin_x: The margin around the box in x direction
            margin_y: The margin around the box in y direction

        Returns:
            Color histogram of the box
        """

        cropped_image = self.crop_img(frame=frame, margin_x=margin_x, margin_y=margin_y)

        if cropped_image is None:
            return None

        # Convert image to HSV (Hue, Saturation, Value)
        hsv_image: ImageType = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)

        h_bins = 10
        s_bins = 10

        # hue varies from 0 to 179, saturation from 0 to 255
        h_ranges = [0, 180]
        s_ranges = [0, 256]

        # Use the 0-th and 1-st channels
        channels = [0, 1]
        color_hist: ImageType = cv2.calcHist(
            [hsv_image], channels, None, [h_bins, s_bins], h_ranges + s_ranges, accumulate=False
        )
        cv2.normalize(color_hist, color_hist, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX)

        return color_hist


def compute_iou(box_1: Box, box_2: Box) -> float:
    """
    Determine the Intersection-over-Union (IoU) of two box objects

    Args:
        box_1: Box object 1
        box_2: Box object 2

    Returns:
        The IoU between box_1 and box_2
    """

    # Determine the coordinates of the intersection rectangle
    x_left = max(box_1.xmin, box_2.xmin)
    x_right = min(box_1.xmax, box_2.xmax)
    y_bottom = min(box_1.ymax, box_2.ymax)
    y_top = max(box_1.ymin, box_2.ymin)

    # Boxes don't overlap at all
    if x_right < x_left or y_bottom < y_top:
        return 0.0

    # The intersection of two axis-aligned bounding boxes
    # is always an axis-aligned bounding box => A(intersect)
    intersection_area = (x_right - x_left) * (y_bottom - y_top)

    # Compute the area of both boxes => A(box_1) and A(box_2(
    box_1_area = (box_1.xmax - box_1.xmin) * (box_1.ymax - box_1.ymin)
    box_2_area = (box_2.xmax - box_2.xmin) * (box_2.ymax - box_2.ymin)

    # Compute the intersection over union by taking the intersection
    # area and dividing it by the sum of box areas:
    # => iou = A(intersect) / ( A(box_1) + A(box_2))
    iou = intersection_area / float(box_1_area + box_2_area - intersection_area)

    # Clamp the iou value
    return max(min(iou, 1.0), 0.0)


def euclidean_distance(box_1: Box, box_2: Box) -> float:
    """
    Determine the euclidean distance between two Box objects

    Args:
        box_1: Box object 1
        box_2: Box object 2

    Returns:
        The euclidean distance between two Box objects
    """
    center1 = box_1.center()
    center2 = box_2.center()

    return float(((center2[0] - center1[0]) ** 2 + (center2[1] - center1[1]) ** 2) ** 0.5)



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for writing Darknet formatted annotations."""
import logging
import os
import random
from pathlib import Path
from shutil import copyfile, rmtree
from typing import List, Optional

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_writer import AnnotationWriter
from mlcvzoo_base.configuration.annotation_handler_config import (
    AnnotationHandlerWriteOutputDarknetAnnotationConfig,
)

logger = logging.getLogger(__name__)


class DarknetAnnotationWriter(AnnotationWriter):
    """
    Writer for generating annotation .txt files that can be used for a training
    of a darknet model, based a given list of annotations in BaseAnnotation format.
    """

    def __init__(
        self,
        darknet_train_set_config: AnnotationHandlerWriteOutputDarknetAnnotationConfig,
        split_size: float,
    ):
        self.darknet_train_set_config = darknet_train_set_config
        self.split_size = split_size

    def write(
        self,
        annotations: List[BaseAnnotation],
    ) -> Optional[str]:
        # Setup directories needed to generate a darknet training set
        output_dir = self.darknet_train_set_config.train_data_set_dir
        jpeg_dir = os.path.join(output_dir, "JPEGImages")
        label_dir = os.path.join(output_dir, "labels")

        if os.path.exists(jpeg_dir):
            rmtree(jpeg_dir)
        if os.path.exists(label_dir):
            rmtree(label_dir)

        os.makedirs(jpeg_dir)
        os.makedirs(label_dir)

        train_txt_path = self.darknet_train_set_config.get_train_file_path()
        test_txt_path = self.darknet_train_set_config.get_test_file_path()

        # open files to put annotation information for training and test data
        train_txt = open(train_txt_path, "w")
        test_txt = open(test_txt_path, "w")

        label_file_extension = ".txt"

        min_test = 1
        min_train = 1
        for _, annotation in enumerate(annotations):
            image_name, image_extension = os.path.splitext(os.path.basename(annotation.image_path))

            with open(
                os.path.join(label_dir, f"{image_name}{label_file_extension}"),
                "a",
            ) as label_txt:
                for bounding_box in annotation.bounding_boxes:
                    w = (bounding_box.box.xmax - bounding_box.box.xmin) / annotation.get_width()
                    h = (bounding_box.box.ymax - bounding_box.box.ymin) / annotation.get_height()
                    x = bounding_box.box.xmin / annotation.get_width() + w / 2
                    y = bounding_box.box.ymin / annotation.get_height() + h / 2
                    label_txt.write("{} {} {} {} {}\n".format(bounding_box.class_id, x, y, w, h))

            img_target = os.path.join(jpeg_dir, f"{image_name}{image_extension}")
            if not os.path.exists(img_target):
                if self.darknet_train_set_config.write_as_symlink:
                    created_relpath = os.path.relpath(annotation.image_path, jpeg_dir)
                    logger.debug("Relative path is '%s'" % created_relpath)

                    symlink_path = Path(img_target)

                    symlink_path.unlink(missing_ok=True)
                    os.symlink(created_relpath, symlink_path)

                    logger.debug("Write image %s as symlink %s", created_relpath, symlink_path)
                else:
                    copyfile(annotation.image_path, img_target)

                    logger.debug("Copy image %s as symlink %s", annotation.image_path, img_target)
            if (random.random() > self.split_size or min_test > 0) and min_train == 0:
                test_txt.write(img_target + "\n")
                min_test -= 1
            else:
                train_txt.write(img_target + "\n")
                min_train -= 1

        train_txt.close()
        test_txt.close()

        return None



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for building Pascal VOC formatted annotations."""
import logging
import os
import xml.etree.ElementTree as ET_xml
from typing import List, Tuple

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_builder import AnnotationBuilder
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.exceptions import ClassMappingNotFoundError
from mlcvzoo_base.data_preparation.utils import ensure_abspath

logger = logging.getLogger(__name__)


class PascalVOCAnnotationBuilder(AnnotationBuilder):
    """
    Super class for defining the methods that are needed to build a BaseAnnotation
    object from a Pascal VOC type XML file.
    """

    def __init__(self, mapper: AnnotationClassMapper, use_difficult: bool) -> None:
        AnnotationBuilder.__init__(self)

        self.mapper = mapper
        self.use_difficult = use_difficult

    def build(
        self,
        image_path: str,
        annotation_path: str,
        image_dir: str,
        annotation_dir: str,
        replacement_string: str,
    ) -> BaseAnnotation:
        (
            annotation_path,
            image_shape,
            bounding_boxes,
            source_image_filename,
        ) = self.__init_from_pascal_voc_xml(xml_file_path=annotation_path)

        if not os.path.isfile(image_path):
            image_path = source_image_filename

        annotation: BaseAnnotation = BaseAnnotation(
            image_path=image_path,
            annotation_path=annotation_path,
            image_shape=image_shape,
            classifications=[],
            bounding_boxes=bounding_boxes,
            segmentations=[],
            image_dir=image_dir,
            annotation_dir=annotation_dir,
            replacement_string=replacement_string,
        )

        try:
            AnnotationBuilder._check_annotation(annotation=annotation)
        except ValueError as value_error:
            logger.exception(
                f"{value_error}, in a future version, the whole annotation will be skipped!"
            )

        annotation = ensure_abspath(annotation=annotation)

        return annotation

    def __init_from_pascal_voc_xml(
        self,
        xml_file_path: str,
    ) -> Tuple[str, Tuple[int, int], List[BoundingBox], str]:
        """
        Parses and generates objects that are needed to build a BaseAnnotation

        Args:
            xml_file_path: Can either be a string to a XML-File or a File-Object of a XML-File
                in VOC-Format

        Returns: the given xml_file_path,
                the shape of the annotated image,
                a list of annotated bounding_boxes and
                the file path of the image, if available in xml file

        """

        bounding_boxes: List[BoundingBox] = list()

        tree = ET_xml.parse(xml_file_path)

        root = tree.getroot()

        width = int(root.find("size").find("width").text)  # type: ignore
        height = int(root.find("size").find("height").text)  # type: ignore
        image_shape = (height, width)

        source_image_filename = ""
        if root.find("path") is not None:
            source_image_filename = (
                root.find("path").text if root.find("path").text is not None else ""  # type: ignore
            )

        for member in root.findall("object"):
            annotation_class_name: str = member[0].text  # type: ignore

            try:
                # map the parsed "class_name" according to the mapping defined in the mapper class
                class_name = self.mapper.map_annotation_class_name_to_model_class_name(
                    class_name=annotation_class_name
                )

                class_id = self.mapper.map_annotation_class_name_to_model_class_id(
                    class_name=annotation_class_name
                )
            except ClassMappingNotFoundError:
                logger.warning(
                    "Could not find a valid class-mapping for class-name '%s'. "
                    "BndBox will be skipped, file = '%s'"
                    % (
                        annotation_class_name,
                        xml_file_path,
                    )
                )
                continue

            # TODO: Handle other properties, like 'occluded', 'truncated'
            # TODO: make it configurable? => add to ClassMapping
            difficult = False
            difficult_tag = member.find("difficult")

            if difficult_tag is not None and difficult_tag.text == "1":
                difficult = True

            box_tag = member.find("bndbox")

            if box_tag is None:
                continue

            bounding_box = BoundingBox(
                box=Box(
                    xmin=int(float(str(box_tag[0].text))),
                    ymin=int(float(str(box_tag[1].text))),
                    xmax=int(float(str(box_tag[2].text))),
                    ymax=int(float(str(box_tag[3].text))),
                ),
                class_identifier=ClassIdentifier(class_id=class_id, class_name=class_name),
                model_class_identifier=ClassIdentifier(class_id=class_id, class_name=class_name),
                difficult=False,
                occluded=False,
                content="",
                score=1.0,
            )

            # TODO: handle occluded as well?
            if difficult and not self.use_difficult:
                logger.debug("Found difficult box, which will be skipped! %s", bounding_box)
                continue

            bounding_boxes.append(bounding_box)

        return xml_file_path, image_shape, bounding_boxes, source_image_filename



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

import os
from dataclasses import dataclass
from subprocess import PIPE, Popen
from typing import Any, Callable, Dict, List, Optional

from mlcvzoo_base.configuration.device_query import ModelTimerDeviceQueryConfig
from mlcvzoo_base.configuration.structs import DeviceQueryTypes


@dataclass
class GpuInfo:
    """
    Datawrapper class for wrapping the information returned by the nvidia-smi command line tool.
    """

    id: int = 0
    uuid: str = ""
    load: float = 0.0
    memory_util: float = 0.0
    memory_total: float = 0.0
    memory_used: float = 0.0
    memory_free: float = 0.0
    driver: str = ""
    name: str = ""
    serial: str = ""
    display_mode: str = ""
    display_active: str = ""
    temperature: float = 0.0


def _cast_to_float(number_str: str) -> float:
    try:
        number = float(number_str)
    except ValueError:
        number = float("nan")
    return number


def _create_gpu_info_setter(
    conversion_func: Callable[[str], Any]
) -> Callable[[GpuInfo, str, Any], None]:
    """
    Args:
        conversion_func: function for converting the input value to the correct
        datatype when setting the respective attribute

    Returns:
        Callable for setting an attribute with name 'attr_name' in
        'gpu_info' to 'conversion_func(value)'
    """

    def _setter(gpu_info: GpuInfo, attr_name: str, value: Any) -> None:
        setattr(gpu_info, attr_name, conversion_func(value))

    return _setter


def get_nvidia_smi_info() -> List[GpuInfo]:
    """
    Query the system for gpu information via nvidia-smi command line tool.

    Returns:
        List of GpuInfo object containing information for the respective gpu.
    """

    query_gpu_dict: Dict[str, Dict[str, Any]] = {
        "index": {
            "attr_name": "id",
            "attr_setter": _create_gpu_info_setter(lambda x: int(x)),
        },
        "uuid": {
            "attr_name": "uuid",
            "attr_setter": _create_gpu_info_setter(lambda x: x),
        },
        "utilization.gpu": {
            "attr_name": "load",
            "attr_setter": _create_gpu_info_setter(lambda x: _cast_to_float(x) / 100),
        },
        "memory.total": {
            "attr_name": "memory_total",
            "attr_setter": _create_gpu_info_setter(lambda x: _cast_to_float(x)),
        },
        "memory.used": {
            "attr_name": "memory_used",
            "attr_setter": _create_gpu_info_setter(lambda x: _cast_to_float(x)),
        },
        "memory.free": {
            "attr_name": "memory_free",
            "attr_setter": _create_gpu_info_setter(lambda x: _cast_to_float(x)),
        },
        "driver_version": {
            "attr_name": "driver",
            "attr_setter": _create_gpu_info_setter(lambda x: x),
        },
        "name": {
            "attr_name": "name",
            "attr_setter": _create_gpu_info_setter(lambda x: x),
        },
        "gpu_serial": {
            "attr_name": "serial",
            "attr_setter": _create_gpu_info_setter(lambda x: x),
        },
        "display_active": {
            "attr_name": "display_active",
            "attr_setter": _create_gpu_info_setter(lambda x: x),
        },
        "display_mode": {
            "attr_name": "display_mode",
            "attr_setter": _create_gpu_info_setter(lambda x: x),
        },
        "temperature.gpu": {
            "attr_name": "temperature",
            "attr_setter": _create_gpu_info_setter(lambda x: _cast_to_float(x)),
        },
    }

    query_gpu_str: str = ",".join(query_gpu_dict.keys())

    try:
        p = Popen(
            [
                "nvidia-smi",
                f"--query-gpu={query_gpu_str}",
                "--format=csv,noheader,nounits",
            ],
            stdout=PIPE,
        )
        stdout, stderr = p.communicate()
    except (RuntimeError, FileNotFoundError):
        # nvidia-smi is not available / has an error, therefore return an empty list
        return []

    smi_output: str = stdout.decode("UTF-8")
    smi_lines: List[str] = smi_output.split(os.linesep)

    num_devices: int = len(smi_lines) - 1
    gpu_infos: List[GpuInfo] = []

    for device_num in range(num_devices):
        line: str = smi_lines[device_num]
        vals: List[str] = line.split(", ")

        gpu_info = GpuInfo()

        for i, key in enumerate(query_gpu_dict):
            attr_name: str = query_gpu_dict[key]["attr_name"]
            value: str = vals[i]
            setter: Callable[[GpuInfo, str, Any], None] = query_gpu_dict[key]["attr_setter"]

            setter(gpu_info, attr_name, value)

        gpu_info.memory_util = float(gpu_info.memory_used) / float(gpu_info.memory_total)
        gpu_infos.append(gpu_info)

    return gpu_infos


def get_device_info(device_query: ModelTimerDeviceQueryConfig) -> Optional[GpuInfo]:
    """
    Determine device information based on the given device query object

    Args:
        device_query: The device query object that should be used
                      to determine the device information

    Returns:
        The device information in form of a GpuInfo object
    """

    gpu_info: Optional[GpuInfo] = None
    if device_query.query_type == DeviceQueryTypes.NVIDIA_SMI.value.upper():
        gpu_info_list: List[GpuInfo] = get_nvidia_smi_info()
        if len(gpu_info_list) != 0:
            if device_query.device_index is not None:
                gpu_info = gpu_info_list[device_query.device_index]
            else:
                gpu_info = gpu_info_list[0]

    return gpu_info


if __name__ == "__main__":
    infos: List[GpuInfo] = get_nvidia_smi_info()

    import pprint

    pprint.pprint(infos)



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for different utility operations during object detection evaluation"""

import logging
import math
import os
from pathlib import Path
from typing import Any, Dict, List, Tuple

import cv2
import numpy as np
from terminaltables import AsciiTable

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.types import ImageType
from mlcvzoo_base.evaluation.object_detection.data_classes import (
    CONFUSION_MATRIX_DICT_TYPE,
    CONFUSION_MATRIX_TYPE,
    METRIC_DICT_TYPE,
    MetricImageInfo,
)
from mlcvzoo_base.evaluation.object_detection.structs import BBoxSizeTypes
from mlcvzoo_base.third_party.efficientdet_pytorch.computer_overlap import (
    compute_overlap,
)

logger = logging.getLogger(__name__)


def get_bbox_size_type(box: Box, image_shape: Tuple[int, int] = (1080, 1440)) -> str:
    """
    Args:
        box: Box object
        image_shape: Tuple of ints, describing the image shape

    Returns:
        a String, the category of the size of the box (small, medium or large)
    """

    bbox_height = (int(box.ymax) - int(box.ymin)) / image_shape[0] * 480
    bbox_width = (int(box.xmax) - int(box.xmin)) / image_shape[1] * 640

    bbox_sqrt_area = math.sqrt(bbox_height * bbox_width)

    small_sqrt_area_limit = 32
    large_sqrt_area_limit = 96

    if bbox_sqrt_area <= small_sqrt_area_limit:
        return str(BBoxSizeTypes.BBOX_SMALL)

    elif small_sqrt_area_limit < bbox_sqrt_area <= large_sqrt_area_limit:
        return str(BBoxSizeTypes.BBOX_MEDIUM)

    else:
        # Any other case is bbox_sqrt_area > large_sqrt_area_limit:
        return str(BBoxSizeTypes.BBOX_LARGE)


def compute_max_bounding_box(
    bounding_box: BoundingBox, gt_bounding_boxes: List[BoundingBox]
) -> Tuple[float, BoundingBox]:
    """
    Determine the ground truth bounding box that has the highest overlap with
    the given (predicted) bounding box

    Args:
        bounding_box: BoundingBox object
        gt_bounding_boxes: List of BoundingBox objects

    Returns:
        A Tuple containing the ground truth bounding box wit the highest overlap and the
        according maximum overlap score (IOU metric)
    """

    gt_bboxes = []
    for gt_bounding_box in gt_bounding_boxes:
        gt_bboxes.append(gt_bounding_box.box.to_list(dst_type=float))

    # Compute Overlap between box and all gt annotations
    # Result in an overlap 2D array of shape (1, K)
    overlaps = compute_overlap(
        np.expand_dims(bounding_box.box.to_list(dst_type=float), axis=0),
        np.asarray(gt_bboxes),
    )
    # get the index of the gt box with the highest overlap (alongside the correct axis)
    max_gt_bbox_index = int(np.argmax(overlaps, axis=1))

    # 0 as index of the first element of 2D-array of shape (1, K)
    max_overlap = float(overlaps[0, max_gt_bbox_index])

    assigned_gt_bounding_box = gt_bounding_boxes[max_gt_bbox_index]

    return max_overlap, assigned_gt_bounding_box


def update_annotation_data(
    classes_id_dict: Dict[int, str],
    all_annotations: List[List[List[BaseAnnotation]]],
    index: int,
    new_annotation: BaseAnnotation,
) -> List[List[List[BaseAnnotation]]]:
    """
    Updates annotation information in the given 'all_annotations' list at the given index.

    Args:
        classes_id_dict: Dictionary of int to String, a class id to name mapping
        all_annotations: stacked List of BaseAnnotations, containing all annotations
        index: int, index of entry (image) which annotations should be updated
        new_annotation: BaseAnnotation object, holding the new annotation information

    Returns: stacked List of BaseAnnotations, the updated list

    """

    logger.warning(
        "DEPRECATED: The method 'update_annotation_data' is deprecated and not used"
        "in any internal code anymore. It will be removed in future versions."
    )

    if not 0 <= index < len(all_annotations):
        logger.debug("WARNING: Index '%s' out of range for all_annotations!", index)
        return all_annotations

    if len(new_annotation.get_bounding_boxes(include_segmentations=True)) > 0:
        # Gather (BoundingBox / Annotation) data for each class-id.
        # This is needed to be able to compute detailed metrics
        # for each class-id.
        for class_id, class_name in classes_id_dict.items():
            # TODO: do we want to add the option to filter for specific class_names?
            if not 0 <= class_id < len(all_annotations[index]):
                logger.debug(
                    "WARNING: Class-ID '%s' out of range for all_annotations[%s]!",
                    class_id,
                    index,
                )
                continue

            # Have to find all bounding-boxes for a specific class-id.
            # In order to have access to the information about the image-path,
            # we need to store an annotation object, rather than a single bounding-box.
            # Despite the fact that for the metric calculation itself the bounding-box
            # information is enough.
            # For purposes like: draw/log all false-positive bounding-boxes for specific
            # images we need to have the information about the image-path
            #   => therefore use BaseAnnotation and not BoundingBox alone
            all_annotations[index][class_id] = []
            for b in new_annotation.get_bounding_boxes(include_segmentations=True):
                if b.class_id == class_id:
                    all_annotations[index][class_id].append(
                        BaseAnnotation(
                            image_path=new_annotation.image_path,
                            annotation_path=new_annotation.annotation_path,
                            image_shape=new_annotation.image_shape,
                            classifications=[],
                            bounding_boxes=[
                                BoundingBox(
                                    class_identifier=b.class_identifier,
                                    model_class_identifier=b.model_class_identifier,
                                    score=b.score,
                                    box=b.box,
                                    difficult=False,
                                    occluded=False,
                                    content="",
                                )
                            ],
                            segmentations=[],
                            image_dir=new_annotation.image_dir,
                            annotation_dir=new_annotation.annotation_dir,
                            replacement_string=new_annotation.replacement_string,
                        )
                    )
    else:
        logger.debug(
            "WARNING: The new_annotation for image %s does not contain "
            "any bounding-boxes! new_annotation: %s",
            new_annotation.image_path,
            new_annotation,
        )

    return all_annotations


def generate_metric_table(
    metrics_dict: METRIC_DICT_TYPE, iou_threshold: float, reduced: bool = False
) -> AsciiTable:
    """
    Generate a 'AsciiTable' object filled with the metrics of a object detection evaluation.
    The columns display the attributes of dataclass
    the mlcvzoo.evaluation.object_detection.data_classes.ODMetrics.

    Args:
        metrics_dict: The dictionary containing the metrics that should be formed into a table
        iou_threshold: The iou-threshold for which the table should be generated
        reduced: Whether to use all available metrics, or only the basic ones

    Returns:
        The generated 'AsciiTable'
    """

    table_data = [
        [
            f"{'class':15s}",
            f"{'TP':6s}",
            f"{'FP':6s}",
            f"{'FN':6s}",
            f"{'CP':6s}",
            f"{'Recall':10s}",
            f"{'Precision':10s}",
            f"{'F1':10s}",
            f"{'AP':8s}",
        ]
    ]
    row_data: List[Any]
    for bbox_size_type in BBoxSizeTypes.get_values_as_list(class_type=BBoxSizeTypes):
        if bbox_size_type != BBoxSizeTypes.BBOX_ALL:
            row_data = [f"SIZE: {bbox_size_type}", "", "", "", "", "", ""]
            table_data.append(row_data)

        for class_name, od_metric in metrics_dict[iou_threshold][bbox_size_type].items():
            if od_metric.COUNT > 0:
                row_data = [
                    class_name,
                    f"{od_metric.TP}",
                    f"{od_metric.FP}",
                    f"{od_metric.FN}",
                    f"{od_metric.COUNT}",
                    f"{od_metric.RC:.4f}",
                    f"{od_metric.PR:.4f}",
                    f"{od_metric.F1:.4f}",
                    f"{od_metric.AP:.4f}",
                ]
            else:
                row_data = [
                    class_name,
                    0,
                    0,
                    0,
                    0,
                    f"{0.0:.4f}",
                    f"{0.0:.4f}",
                    f"{0.0:.4f}",
                    f"{0.0:.4f}",
                ]

            table_data.append(row_data)

        if bbox_size_type != "l":
            row_data = ["", "", "", "", "", "", ""]
            table_data.append(row_data)

    table = AsciiTable(table_data)
    table.inner_footing_row_border = False

    return table


def generate_fn_fp_confusion_matrix_table(
    confusion_matrix: CONFUSION_MATRIX_TYPE, classes_id_dict: Dict[int, str]
) -> AsciiTable:
    """
    Generate a 'AsciiTable' object that represents the confusion matrix
    of an objection detection evaluation output. The confusion matrix indicates
    which false negative bounding boxes of a certain class could be matches to
    false positive bounding boxes of another class.

    Args:
        confusion_matrix: The confusion Matrix as 2D List
        classes_id_dict: Dictionary that matches the class IDs in the confusion matrix
                         to class names

    Returns:
        The generated 'AsciiTable'
    """

    logger.warning(
        "DEPRECATED: The method 'generate_fn_fp_confusion_matrix_table' is deprecated"
        "and will be removed in future versions!"
    )

    table_data = [[""]]

    for class_id in range(0, len(confusion_matrix)):
        table_data[0].append(classes_id_dict[class_id])

    row_data: List[Any]
    for row_index in range(0, len(confusion_matrix)):
        row_data = [classes_id_dict[row_index]]
        for column_index in range(0, len(confusion_matrix)):
            row_data.append(f"{confusion_matrix[row_index][column_index]}")
        table_data.append(row_data)

    table = AsciiTable(table_data)
    table.inner_footing_row_border = False

    return table


def generate_fn_fp_confusion_matrix_table_from_dict(
    confusion_matrix: CONFUSION_MATRIX_DICT_TYPE,
) -> AsciiTable:
    """
    Generate a 'AsciiTable' object that represents the confusion matrix
    of an objection detection evaluation output. The confusion matrix indicates
    which false negative bounding boxes of a certain class could be matches to
    false positive bounding boxes of another class.

    Args:
        confusion_matrix: The confusion Matrix as 2D Dict

    Returns:
        The generated 'AsciiTable'
    """

    table_data = [[""]]

    for class_identifier_str in confusion_matrix.keys():
        table_data[0].append(class_identifier_str)

    row_data: List[Any]
    for row_class_identifier_str in confusion_matrix.keys():
        row_data = [row_class_identifier_str]
        for column_class_identifier_str in confusion_matrix[row_class_identifier_str].keys():
            row_data.append(
                f"{confusion_matrix[row_class_identifier_str][column_class_identifier_str]}"
            )
        table_data.append(row_data)

    table = AsciiTable(table_data)
    table.inner_footing_row_border = False

    return table


def __create_tf_image(
    image_path: str, metric_image_info: MetricImageInfo, false_positive_image_size: int
) -> ImageType:
    img = cv2.imread(image_path)

    fn_color = (255, 0, 0)  # => red colour for false negative boxes
    fp_color = (0, 0, 0)  # => black colour for false positive boxes
    gt_color = (255, 255, 255)  # => white colour for ground truth data

    if (
        metric_image_info.false_positive_annotation is not None
        or metric_image_info.false_negative_annotation is not None
    ) and metric_image_info.ground_truth_annotation is not None:
        for bounding_box in metric_image_info.ground_truth_annotation.get_bounding_boxes(
            include_segmentations=True
        ):
            cv2.rectangle(
                img,
                (bounding_box.box.xmin, bounding_box.box.ymin),
                (bounding_box.box.xmax, bounding_box.box.ymax),
                gt_color,
                6,
            )

    if metric_image_info.false_positive_annotation is not None:
        for bounding_box in metric_image_info.false_positive_annotation.get_bounding_boxes(
            include_segmentations=True
        ):
            cv2.rectangle(
                img,
                (bounding_box.box.xmin, bounding_box.box.ymin),
                (bounding_box.box.xmax, bounding_box.box.ymax),
                fp_color,
                2,
            )

    if metric_image_info.false_negative_annotation is not None:
        for bounding_box in metric_image_info.false_negative_annotation.get_bounding_boxes(
            include_segmentations=True
        ):
            cv2.rectangle(
                img,
                (bounding_box.box.xmin, bounding_box.box.ymin),
                (bounding_box.box.xmax, bounding_box.box.ymax),
                fn_color,
                2,
            )

    # Convert to RGB for tensorboard
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

    # resize to the configured image size
    height, width, _ = img.shape
    scale_factor = false_positive_image_size / width
    img = cv2.resize(
        img,
        (int(width * scale_factor), int(height * scale_factor)),
        interpolation=cv2.INTER_AREA,
    )

    return np.asarray(img).astype(np.uint8)


def generate_img_id_map(
    image_path: str, img_directory_id_dict: Dict[str, int]
) -> Tuple[int, Dict[str, int]]:
    """
    Generates an index (id) for the given image_path.

    Args:
        image_path: String, path to an image
        img_directory_id_dict: Dictionary of image paths to index in directory list

    Returns: a Tuple of image index and (updated) img_directory_id_dict
    """
    # TODO: What happens when two images get the same index?
    #  It could be possible that due to an extension of the dict two images could get the same index,
    #  because the indices after the added and sorted image are increased by 1.
    #  Therefore all other entries of the subsequent images (in sorted order) are wrong.

    image_dir = os.path.dirname(image_path)
    file_extension = os.path.splitext(image_path)[1]

    if image_path not in img_directory_id_dict:
        path_generator = Path(image_dir).glob(f"**/*{file_extension}")

        file_list = [str(p) for p in path_generator]

        file_list.sort()

        for index, file_path in enumerate(file_list):
            img_directory_id_dict[file_path] = index

    return img_directory_id_dict[image_path], img_directory_id_dict



# Copyright 2022 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

from typing import Optional

import related
from attr import define
from config_builder import BaseConfigClass

from mlcvzoo_base.configuration.structs import DeviceQueryTypes


@define
class ModelTimerDeviceQueryConfig(BaseConfigClass):
    device_index: Optional[int] = related.ChildField(cls=int)
    query_type: str = related.StringField()

    def check_values(self) -> bool:
        return (
            self.device_index is None
            or isinstance(self.device_index, int)
            and self.query_type in [d.value.upper() for d in DeviceQueryTypes]
        )



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for building MOT formatted annotations."""
import logging
from typing import Dict, List, Tuple

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_builder import AnnotationBuilder
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.exceptions import ForbiddenClassError
from mlcvzoo_base.configuration.structs import ObjectDetectionBBoxFormats
from mlcvzoo_base.data_preparation.structs import MOTChallengeFormats
from mlcvzoo_base.data_preparation.utils import ensure_abspath

logger = logging.getLogger(__name__)


class MOTAnnotationBuilder(AnnotationBuilder):
    """
    Class for defining the methods that are needed to build a BaseAnnotation
    object from an MOT annotation file (.txt).

    REMARK: The MOT Challenge only considers BoundingBoxes.
    """

    def __init__(
        self,
        image_shape: Tuple[int, int],  # HEIGHT, WIDTH
        image_id: int,
        annotation_line_dict: Dict[int, List[str]],
        mapper: AnnotationClassMapper,
        mot_class_mapper: AnnotationClassMapper,
        mot_format: str,
        ground_truth: bool,
    ) -> None:
        AnnotationBuilder.__init__(self)

        self.image_shape = image_shape
        self.image_id = image_id
        self.annotation_line_dict = annotation_line_dict
        self.mapper = mapper
        self.mot_class_mapper = mot_class_mapper
        self.mot_format = mot_format
        self.ground_truth = ground_truth

    def build(
        self,
        image_path: str,
        annotation_path: str,
        image_dir: str,
        annotation_dir: str,
        replacement_string: str,
    ) -> BaseAnnotation:
        """
        Build an annotation from the given annotation lines of the annotation_line_dict
        attribute, which have been extracted form an MOT annotation file.

        Args:
            image_path: String, points to an annotated image
            annotation_path: String, points to the respective annotation
            image_dir: String, points to the dir where the annotated image is stored
            annotation_dir: String, points to the dir where the respective annotation is stored
            replacement_string: String, part of the paths that is a placeholder

        Returns:
            A BaseAnnotation object
        """

        bounding_boxes = self.__parse_bounding_boxes()

        annotation: BaseAnnotation = BaseAnnotation(
            image_path=image_path,
            annotation_path=annotation_path,
            image_shape=self.image_shape,
            classifications=[],
            bounding_boxes=bounding_boxes,
            image_dir=image_dir,
            annotation_dir=annotation_dir,
            replacement_string=replacement_string,
        )

        annotation = ensure_abspath(annotation=annotation)

        try:
            AnnotationBuilder._check_annotation(annotation=annotation)
        except ValueError as value_error:
            logger.exception(
                f"{value_error}, in a future version, the whole annotation will be skipped!"
            )

        return annotation

    def __parse_bounding_boxes(
        self,
    ) -> List[BoundingBox]:
        """
        Parse bounding boxes from all annotation lines that are associated with the
        image-id attribute ob this MOTAnnotationBuilder instance.

        Format for MOT2015 is:
        <frame>, <id>, <bb_left>, <bb_top>, <bb_width>, <bb_height>, <conf>, <3D-X>, <3D-Y>, <3D-Z>

        Format for MOT201617 and MOT2020 is:
        <frame>, <id>, <bb_left>, <bb_top>, <bb_width>, <bb_height>, <conf>, <class>, <visibility>

        Explanation of fields, from the respective papers:
        - frame: Number of the frame in the image sequence
        - id: bounding box ID in a trajectory
        - confidence score: in ground truth data handled as a flag (0/1) for consideration of boxed
        - 3D world coordinates: describe the position of a pedestrian's feet, "in the case of
                                2D tracking, these values will be ignored and can be left at -1"
        - class id: introduced in MOT201617, before only pedestrians were considered
        - visibility: indicates the degree of visibility of an object

        Note from the instruction page at https://motchallenge.net/instructions/ :
        "All frame numbers, target IDs and bounding boxes are 1-based"

        REMARK: <id> field is not used in object detection tasks. In case you want to
                pursue tracking this functionality needs to be added.

        Details about the annotation format can be found here:

        Documentation regarding the MOT challenge submission format:
        https://motchallenge.net/instructions/

        The format for the MOT challenge 2015:
        https://arxiv.org/pdf/1504.01942.pdf (section 3.4 "Data Format")

        The format for the MOT challenge 2016:
        https://arxiv.org/pdf/1603.00831.pdf (section 3.4 "Data Format")

        Returns:
            A list of BoundingBox objects for image of the MOTAnnotationBuilder instance

        Raises:
            ClassMappingNotFoundError: when the class_name can not be mapped to a model class
            name.
        """

        bounding_boxes: List[BoundingBox] = []

        if self.image_id not in self.annotation_line_dict:
            return bounding_boxes

        for mot_annotation_line in self.annotation_line_dict[self.image_id]:
            # Note: annotation is given in a .txt file where each line equals to one bounding box.
            # These lines are contained in self.mot_annotations as a list of strings.
            # So each line has to be read out, split along the ",", filtered by image_id
            # and the corresponding bounding box values parsed to integer values.
            mot_annotation_elements = mot_annotation_line.split(",")

            # The mot class mapping is static. Therefore, first translate the given
            # class-id to a mot class-name and then map this class-name to the according
            # class-id and class-name of the model.
            if self.mot_format == MOTChallengeFormats.MOT15.value:
                mot_class_id = 1
            else:
                mot_class_id = int(mot_annotation_elements[7])

            mot_class_name = self.mot_class_mapper.map_annotation_class_id_to_model_class_name(
                class_id=mot_class_id
            )

            try:
                class_id = self.mapper.map_annotation_class_name_to_model_class_id(
                    class_name=mot_class_name
                )
                class_name = self.mapper.map_annotation_class_name_to_model_class_name(
                    class_name=mot_class_name
                )
            except ForbiddenClassError:
                continue

            # Note:
            #  - in GT data "confidence score" indicates whether an entry is considered for
            #    training
            #  - Whereas in pre-annotated data it indicates how confident the detector was
            #    that this instance is of the specified class [1,100]
            score = float(mot_annotation_elements[6])
            if self.ground_truth and score == 0.0:
                # Skip annotation line if mot annotation file contains ground truth data.
                # Then score value serves as flag for considering an annotation during
                # training (1=yes) or not (0=no).
                continue

            bounding_boxes.append(
                BoundingBox(
                    class_identifier=ClassIdentifier(
                        class_id=class_id,
                        class_name=class_name,
                    ),
                    score=score,
                    difficult=False,
                    occluded=False,
                    content="",
                    box=Box.init_format_based(  # NOTE: MOT bbox is format XYWH
                        box_list=(
                            int(float(mot_annotation_elements[2])),
                            int(float(mot_annotation_elements[3])),
                            int(float(mot_annotation_elements[4])),
                            int(float(mot_annotation_elements[5])),
                        ),
                        box_format=ObjectDetectionBBoxFormats.XYWH,
                        src_shape=self.image_shape,
                    ),
                )
            )
        return bounding_boxes



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for handling implicit replacements"""

__author__ = "Oliver Bredtmann"
__license__ = "Open Logistics License 1.0"
__email__ = "Oliver.Bredtmann@dbschenker.com"

import os


class ImplicitReplacement:
    """
    Save replacement of
        - directories
        - strings
    """

    ALL = "all"
    FIRST = "first"
    LAST = "last"
    HOW = [ALL, FIRST, LAST]

    @staticmethod
    def replace_directory_in_path(
        file_path: str, replacement_key: str, replacement_value: str, how: str = ALL
    ) -> str:
        """
        Replaces directory in the given path.

        Args:
            file_path: String, a path to a file
            replacement_key: String, defining what is about to be replaced
            replacement_value: String, defining what the replacement looks like
            how: String, one of ImplicitReplacement enums [ALL, FIRST, LAST]

        Returns: String, the updated file path

        """

        assert how in ImplicitReplacement.HOW

        path_list = os.path.normpath(file_path).split(os.sep)

        if how == ImplicitReplacement.ALL:
            path_list = list(
                map(
                    lambda x: x if x != replacement_key else replacement_value,
                    path_list,
                )
            )

        elif how == ImplicitReplacement.FIRST:
            for idx in range(len(path_list)):
                if path_list[idx] == replacement_key:
                    path_list[idx] = replacement_value
                    break

        elif how == ImplicitReplacement.LAST:
            for idx in list(range(len(path_list)))[::-1]:
                if path_list[idx] == replacement_key:
                    path_list[idx] = replacement_value
                    break

        new_path = os.sep.join(path_list)

        return new_path

    @staticmethod
    def replace_string_in_path(
        file_path: str, value: str, replacement_value: str, how: str = ALL
    ) -> str:
        """
        Replaces a string in the given path.

        Args:
            file_path: String, a path to a file
            value: String, defining what is about to be replaced
            replacement_value: String, defining what the replacement looks like
            how:

        Returns: String, one of ImplicitReplacement enums [ALL, FIRST, LAST]

        """

        assert how in ImplicitReplacement.HOW

        new_path = file_path
        if how == ImplicitReplacement.ALL:
            new_path = file_path.replace(value, replacement_value)

        elif how == ImplicitReplacement.FIRST:
            new_path = file_path.replace(value, replacement_value, 1)

        elif how == ImplicitReplacement.LAST:
            new_path = (file_path[::-1].replace(value[::-1], replacement_value[::-1], 1))[::-1]

        return new_path
